import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
import pypesto

from typing import Optional, Tuple


def optimizer_convergence(result: pypesto.Result,
                          ax: Optional[plt.Axes] = None,
                          xscale: str = 'symlog',
                          yscale: str = 'log',
                          size: Tuple[float] = (18.5, 10.5)) -> plt.Axes:
    """
    Scatter plot of function values and gradient values at the end of
    optimization. Optimizer exit-message is encoded by color. Can help
    identifying convergence issues in optimization and guide tolerance
    refinement etc.

    Parameters
    ----------

    result:
        Optimization result obtained by 'optimize.py'

    ax:
        Axes object to use.

    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    xscale:
        Scale for x-axis

    yscale:
        Scale for y-axis

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """
    if ax is None:
        ax = plt.subplots(figsize=size)[1]

    fvals = result.optimize_result.get_for_key('fval')
    grad_norms = [
        np.linalg.norm(
            result.problem.get_reduced_vector(grad,
                                              result.problem.x_free_indices),
            2
        )
        if grad is not None else np.NaN
        for grad in result.optimize_result.get_for_key('grad')
    ]
    msgs = result.optimize_result.get_for_key('message')
    conv_data = pd.DataFrame({
        'fval': fvals,
        'gradient norm': grad_norms,
        'exit message': msgs
    })
    sns.scatterplot(x='fval', y='gradient norm', hue='exit message',
                    data=conv_data, ax=ax)
    ax.set_yscale(yscale)
    ax.set_xscale(xscale)
    return ax



"""
pyPESTO
=======

Parameter Estimation TOolbox for python.
"""

# make version available
from .version import __version__

# import basic objects into global namespace
from .objective import (
    AmiciObjective,
    CsvHistory,
    Hdf5History,
    History,
    HistoryBase,
    HistoryOptions,
    MemoryHistory,
    NegLogPriors,
    Objective,
    ObjectiveBase,
    OptimizerHistory,
    FD,
    FDDelta,
)
from .problem import Problem
from .result import (
    OptimizeResult,
    ProfileResult,
    Result,
    SampleResult,
)

# import simple modules as submodules
from . import engine
from . import logging
from . import startpoint
from . import store
from . import visualize

logging.log()



import numpy as np


class ProfilerResult(dict):
    """
    The result of a profiler run. The standardized return return value from
    pypesto.profile, which can either be initialized from an OptimizerResult
    or from an existing ProfilerResult (in order to extend the computation).

    Can be used like a dict.

    Attributes
    ----------
    x_path:
        The path of the best found parameters along the profile
        (Dimension: n_par x n_profile_points)
    fval_path:
        The function values, fun(x), along the profile.
    ratio_path:
        The ratio of the posterior function along the profile.
    gradnorm_path:
        The gradient norm along the profile.
    exitflag_path:
        The exitflags of the optimizer along the profile.
    time_path:
        The computation time of the optimizer runs along the profile.
    time_total:
        The total computation time for the profile.
    n_fval:
        Number of function evaluations.
    n_grad:
        Number of gradient evaluations.
    n_hess:
        Number of Hessian evaluations.
    message:
        Textual comment on the profile result.

    Notes
    -----

    Any field not supported by the profiler or the profiling optimizer is
    filled with None. Some fields are filled by pypesto itself.
    """

    def __init__(self,
                 x_path: np.ndarray,
                 fval_path: np.ndarray,
                 ratio_path: np.ndarray,
                 gradnorm_path: np.ndarray = None,
                 exitflag_path: np.ndarray = None,
                 time_path: np.ndarray = None,
                 time_total: float = 0.,
                 n_fval: int = 0,
                 n_grad: int = 0,
                 n_hess: int = 0,
                 message: str = None):
        super().__init__()

        # initialize profile path
        x_shape = x_path.shape
        if len(x_shape) == 1:
            self.x_path = np.zeros((x_shape[0], 1))
            self.x_path[:, 0] = x_path[:]
        else:
            self.x_path = np.zeros((x_shape[0], x_shape[1]))
            self.x_path[:, :] = x_path[:, :]

        self.fval_path = np.asarray(fval_path)
        self.ratio_path = np.asarray(ratio_path)
        self.gradnorm_path = np.asarray(gradnorm_path) \
            if gradnorm_path is not None else None
        self.exitflag_path = np.asarray(exitflag_path)
        self.time_path = np.asarray(time_path)
        self.time_total = time_total
        self.n_fval = n_fval
        self.n_grad = n_grad
        self.n_hess = n_hess
        self.message = message

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def append_profile_point(self,
                             x: np.ndarray,
                             fval: float,
                             ratio: float,
                             gradnorm: float = np.nan,
                             time: float = np.nan,
                             exitflag: float = np.nan,
                             n_fval: int = 0,
                             n_grad: int = 0,
                             n_hess: int = 0) -> None:
        """
        This function appends a new point to the profile path.

        Parameters
        ----------
        x:
            The parameter values.
        fval:
            The function value at `x`.
        ratio:
            The ratio of the function value at `x` by the optimal function
            value.
        gradnorm:
            The gradient norm at `x`.
        time:
            The computation time to find `x`.
        exitflag:
            The exitflag of the optimizer (useful if an optimization was
            performed to find `x`).
        n_fval:
            Number of function evaluations performed to find `x`.
        n_grad:
            Number of gradient evaluations performed to find `x`.
        n_hess:
            Number of Hessian evaluations performed to find `x`.
        """

        # short function to append to numpy vectors
        def append_to_vector(field_name, val):
            field_new = np.zeros(self[field_name].size + 1)
            field_new[0:-1] = self[field_name]
            field_new[-1] = val
            self[field_name] = field_new

        # write profile path
        x_new = np.zeros((self.x_path.shape[0], self.x_path.shape[1] + 1))
        x_new[:, 0:-1] = self.x_path
        x_new[:, -1] = x
        self.x_path = x_new

        # append to other paths
        append_to_vector("fval_path", fval)
        append_to_vector("ratio_path", ratio)
        append_to_vector("gradnorm_path", gradnorm)
        append_to_vector("exitflag_path", exitflag)
        append_to_vector("time_path", time)

        # increment the time and f_eval counters
        self.time_total += time
        self.n_fval += n_fval
        self.n_grad += n_grad
        self.n_hess += n_hess

    def flip_profile(self) -> None:
        """
        This function flips the profiling direction (left-right)
        Profiling direction needs to be changed once (if the profile is new),
        or twice if we append to an existing profile.

        All profiling paths are flipped in-place.
        """

        self.x_path = np.fliplr(self.x_path)
        self.fval_path = np.flip(self.fval_path)
        self.ratio_path = np.flip(self.ratio_path)
        self.gradnorm_path = np.flip(self.gradnorm_path)
        self.exitflag_path = np.flip(self.exitflag_path)
        self.time_path = np.flip(self.time_path)



"""Constants related to visualization methods."""

from typing import Tuple, Union

LEN_RGB = 3  # number of elements in an RGB color
LEN_RGBA = 4  # number of elements in an RGBA color
RGB = Tuple[(float,) * LEN_RGB]  # typing of an RGB color
RGBA = Tuple[(float,) * LEN_RGBA]  # typing of an RGBA color
RGB_RGBA = Union[RGB, RGBA]  # typing of an RGB or RGBA color
RGBA_MIN = 0  # min value for an RGBA element
RGBA_MAX = 1  # max value for an RGBA element
RGBA_ALPHA = 3  # zero-indexed fourth element in RGBA
RGBA_WHITE = (RGBA_MAX, RGBA_MAX, RGBA_MAX, RGBA_MAX)  # white as an RGBA color
RGBA_BLACK = (RGBA_MIN, RGBA_MIN, RGBA_MIN, RGBA_MAX)  # black as an RGBA color



"""
Visualize
=========

pypesto comes with various visualization routines. To use these,
import pypesto.visualize.
"""

from .reference_points import (
    ReferencePoint,
    create_references,
)
from .clust_color import (
    assign_clusters,
    assign_clustered_colors,
    assign_colors,
    delete_nan_inf,
)
from .misc import (
    process_result_list,
    process_offset_y,
    process_y_limits,
)
from .waterfall import (
    waterfall,
    waterfall_lowlevel,
)
from .parameters import (
    parameters,
    parameters_lowlevel,
    parameter_hist,
)
from .optimizer_history import (
    optimizer_history,
    optimizer_history_lowlevel,
)
from .optimization_stats import (
    optimization_run_properties_per_multistart,
    optimization_run_property_per_multistart,
    optimization_run_properties_one_plot,
)
from .optimizer_convergence import optimizer_convergence
from .profiles import (
    profiles,
    profiles_lowlevel,
    profile_lowlevel,
)
from .profile_cis import profile_cis
from .sampling import (
    sampling_fval_traces,
    sampling_parameter_traces,
    sampling_scatter,
    sampling_1d_marginals,
    sampling_parameter_cis,
    sampling_prediction_trajectories,
)
from .ensemble import ensemble_identifiability
from .dimension_reduction import (projection_scatter_umap,
                                  projection_scatter_umap_original,
                                  projection_scatter_pca,
                                  ensemble_crosstab_scatter_lowlevel,
                                  ensemble_scatter_lowlevel)



"""
Visualization of the model fit after optimization.

Currently only for PEtab problems.
"""
import numpy as np
import matplotlib.pyplot as plt
import amici.petab_import as petab_import
import petab
import amici.plotting
from ..problem import Problem
from amici.petab_simulate import simulate_petab
from petab.visualize import plot_problem
from typing import Union, Sequence
from ..result import Result
from amici.petab_objective import rdatas_to_simulation_df

AmiciModel = Union['amici.Model', 'amici.ModelPtr']


def visualize_optimized_model_fit(petab_problem: petab.Problem,
                                  result: Union[Result, Sequence[Result]],
                                  **kwargs
                                  ):
    """
    Visualize the optimized model fit of a PEtab problem.
    Function calls the PEtab visualization file of the petab_problem and
    visualizes the fit of the optimized parameter. Common additional
    argument is `subplot_dir` to specify the directory each subplot is
    saved to. Further keyword arguments are delegated to
    petab.visualize.plot_with_vis_spec, see there for more information.

    Parameters
    ----------
    petab_problem:
        The :py:class:`petab.Problem` that was optimized.
    result:
        The result object from optimization.

    Returns
    -------
    ax: Axis object of the created plot.
    None: In case subplots are saved to file
    """
    if petab_problem is not None:
        if petab is None:
            raise

    problem_parameters = \
        dict(zip(petab_problem.parameter_df.index,
                 result.optimize_result.list[0]['x']))

    amici_model = petab_import.import_petab_problem(
        petab_problem,
        model_output_dir=kwargs.pop('model_output_dir', None),
        force_compile=kwargs.pop('force_compile', False))

    res = simulate_petab(petab_problem, amici_model=amici_model,
                         scaled_parameters=True,
                         problem_parameters=problem_parameters,
                         solver=kwargs.pop('amici_solver', None))

    sim_df = rdatas_to_simulation_df(res["rdatas"],
                                     amici_model,
                                     petab_problem.measurement_df)

    # function to call, to plot data and simulations
    ax = plot_problem(petab_problem=petab_problem,
                      simulations_df=sim_df,
                      **kwargs
                      )
    return ax


def time_trajectory_model(
        result: Union[Result, Sequence[Result]],
        problem: Problem = None,
        # TODO: conditions: Union[str, Sequence[str]] = None,
        timepoints: Union[np.ndarray, Sequence[np.ndarray]] = None,
        n_timepoints: int = 1000,
        start_index: int = 0,
        state_ids: Union[str, Sequence[str]] = None,
        state_names: Union[str, Sequence[str]] = None,
        observable_ids: Union[str, Sequence[str]] = None,):
    """
    Visualize the time trajectory of the model with given timepoints.
    It does this by calling the amici plotting routines.

    Parameters
    ----------
    result:
        The result object from optimization.
    problem:
        A pypesto problem. Default is 'None' in which case result.problem is
        used. Needed in case the result is loaded from hdf5.
    timepoints:
        Array of timepoints, at which the trajectory will be plotted.
    n_timepoints:
        Number of timepoints to be plotted between 0 and last measurement of
        the model. Only used when timepoints==None.
    start_index:
        Index of Optimization run to be plotted. Default is best start.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes: `matplotlib.axes.Axes` object of the plot.
    """

    if problem is None:
        problem = result.problem
    # add timepoints as needed
    if timepoints is None:
        end_time = max(problem.objective.edatas[0].getTimepoints())
        timepoints = np.linspace(start=0,
                                 stop=end_time,
                                 num=n_timepoints)
    obj = problem.objective.set_custom_timepoints(
        timepoints_global=timepoints)

    # evaluate objective with return dic = True to get data
    parameters = result.optimize_result.list[start_index]['x']
    # reduce vector to only include free indices. Needed downstream.
    parameters = problem.get_reduced_vector(parameters)
    ret = obj(parameters, mode='mode_fun',
              sensi_orders=(0,), return_dict=True)

    if state_ids == [] and state_names == []:
        axes = _time_trajectory_model_without_states(
            model=problem.objective.amici_model,
            rdatas=ret['rdatas'],
            observable_ids=observable_ids
        )
    else:
        axes = _time_trajectory_model_with_states(
            model=problem.objective.amici_model,
            rdatas=ret['rdatas'],
            state_ids=state_ids,
            state_names=state_names,
            observable_ids=observable_ids
        )

    return axes


def _time_trajectory_model_with_states(
        model: AmiciModel,
        rdatas: Union['amici.ReturnData', Sequence['amici.ReturnData']],
        state_ids: Sequence[str],
        state_names: Sequence[str],
        observable_ids: Union[str, Sequence[str]]):
    """
    Helper function for time_trajectory_model. Visualizes both states and
    observables.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observable Ids to be plotted.

    Returns
    -------
    axes: `matplotlib.axes.Axes` object of the plot.
    """
    # if state_name, state_id or observable_id is not None, get indices
    # for these the AMICI plotting functions default to all indices if
    # `None` is specified.
    state_indices_by_id = []
    state_indices_by_name = []
    if state_ids is not None:
        state_indices_by_id = [
            model.getStateIds().index(state_id)
            for state_id in state_ids]
    if state_names is not None:
        state_indices_by_name = [
            model.getStateNames().index(state_name)
            for state_name in state_names]
    state_indices = list(set(state_indices_by_id + state_indices_by_name))
    if state_indices == []:
        state_indices = None

    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.getObservableIds().index(obs_id)
            for obs_id in observable_ids]

    fig, axes = plt.subplots(len(rdatas), 2)
    # enforce two dimensions in case there is only one condition
    axes = np.atleast_2d(axes)

    for i_cond, rdata in enumerate(rdatas):
        amici.plotting.plotStateTrajectories(
            rdata=rdata,
            state_indices=state_indices,
            ax=axes[i_cond, 0],
            model=model)
        amici.plotting.plotObservableTrajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond, 1],
            model=model)
    return axes


def _time_trajectory_model_without_states(
        model: AmiciModel,
        rdatas: Union['amici.ReturnData', Sequence['amici.ReturnData']],
        observable_ids: Union[str, Sequence[str]]):
    """
    Helper function for time_trajectory_model. Visualizes both states and
    observables.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes: `matplotlib.axes.Axes` object of the plot.
    """
    # if observable_id's is not None, get indices for these
    # the AMICI plotting functions default to all indices if `None` is
    # specified.
    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.getObservableIds().index(obs_id)
            for obs_id in observable_ids]

    fig, axes = plt.subplots(len(rdatas))

    for i_cond, rdata in enumerate(rdatas):
        amici.plotting.plotObservableTrajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond],
            model=model)
    return axes



from typing import List
import abc

from .task import Task


class Engine(abc.ABC):
    """
    Abstract engine base class.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Execute tasks.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        raise NotImplementedError(
            "This engine is not intended to be called.")



"""
This is for (string) constants used in the ensemble module.
"""


from enum import Enum
from typing import Union

from ..predict.constants import (  # noqa: F401
    OUTPUT,
    OUTPUT_IDS,
    OUTPUT_SENSI,
    PARAMETER_IDS,
    TIMEPOINTS,
)


MODE_FUN = 'mode_fun'  # mode for function values

PREDICTOR = 'predictor'
PREDICTION_ID = 'prediction_id'
PREDICTION_RESULTS = 'prediction_results'
PREDICTION_ARRAYS = 'prediction_arrays'
PREDICTION_SUMMARY = 'prediction_summary'

HISTORY = 'history'
OPTIMIZE = 'optimize'
SAMPLE = 'sample'

MEAN = 'mean'
MEDIAN = 'median'
STANDARD_DEVIATION = 'std'
PERCENTILE = 'percentile'
SUMMARY = 'summary'
WEIGHTED_SIGMA = 'weighted_sigma'

X_NAMES = 'x_names'
NX = 'n_x'
X_VECTOR = 'x_vectors'
NVECTORS = 'n_vectors'
VECTOR_TAGS = 'vector_tags'
ENSEMBLE_TYPE = 'ensemble_type'
PREDICTIONS = 'predictions'

LOWER_BOUND = 'lower_bound'
UPPER_BOUND = 'upper_bound'
PREEQUILIBRATION_CONDITION_ID = 'preequilibrationConditionId'
SIMULATION_CONDITION_ID = 'simulationConditionId'

CSV = 'csv'  # return file format
H5 = 'h5'  # return file format
TIME = 'time'  # column name in returned csv

COLOR_HIT_BOTH_BOUNDS = [0.6, 0., 0., 0.9]
COLOR_HIT_ONE_BOUND = [0.95, 0.6, 0., 0.9]
COLOR_HIT_NO_BOUNDS = [0., 0.8, 0., 0.9]


class EnsembleType(Enum):
    ensemble = 1
    sample = 2
    unprocessed_chain = 3


def get_percentile_label(percentile: Union[float, int, str]) -> str:
    """Convert a percentile to a label.

    Labels for percentiles are used at different locations (e.g. ensemble
    prediction code, and visualization code). This method ensures that the same
    percentile is labeled identically everywhere.

    The percentile is rounded to two decimal places in the label representation
    if it is specified to more decimal places. This is for readability in
    plotting routines, and to avoid float to string conversion issues related
    to float precision.

    Parameters
    ----------
    percentile:
        The percentile value that will be used to generate a label.

    Returns
    -------
    The label of the (possibly rounded) percentile.
    """
    if isinstance(percentile, str):
        percentile = float(percentile)
        if percentile == round(percentile):
            percentile = round(percentile)
    if isinstance(percentile, float):
        percentile_str = f'{percentile:.2f}'
        # Add `...` to the label if the percentile value changed after rounding
        if float(percentile_str) != percentile:
            percentile_str += '...'
        percentile = percentile_str
    return f'{PERCENTILE} {percentile}'



"""Assign startpoints for a given problem."""

import numpy as np
from typing import Union

from .base import StartpointMethod
from ..problem import Problem
from ..objective import ObjectiveBase


def assign_startpoints(
    n_starts: int,
    startpoint_method: Union[StartpointMethod, bool],
    problem: Problem,
    startpoint_resample: bool,
) -> np.ndarray:
    """Assign startpoints for a given problem.

    This is the main method called e.g. by `pypesto.optimize.minimize`.

    Parameters
    ----------
    n_starts:
        Number of startpoints to generate.
    startpoint_method:
        Startpoint generation method to use.
    problem:
        Underlying problem specifying e.g. dimensions and bounds.
    startpoint_resample:
        Whether to evaluate function values at proposed startpoints, and
        resample ones having non-finite values until all startpoints have
        finite value.

    Returns
    -------
    startpoints:
        Startpoints, shape (n_starts, n_par).
    """
    # check if startpoints needed
    if startpoint_method is False:
        # fill with dummies
        startpoints = np.zeros(n_starts, problem.dim)
        startpoints[:] = np.nan
        return startpoints

    x_guesses = problem.x_guesses
    dim = problem.lb.size

    # number of required startpoints
    n_guessed_points = x_guesses.shape[0]
    n_required_points = n_starts - n_guessed_points

    if n_required_points <= 0:
        return x_guesses[:n_starts, :]

    # apply startpoint method
    x_sampled = startpoint_method(
        n_starts=n_required_points,
        lb=problem.lb_init,
        ub=problem.ub_init,
        objective=problem.objective,
    )

    # put together
    startpoints = np.zeros((n_starts, dim))
    startpoints[0:n_guessed_points, :] = x_guesses
    startpoints[n_guessed_points:n_starts, :] = x_sampled

    # resample and order startpoints
    if startpoint_resample:
        startpoints = resample_startpoints(
            startpoints=startpoints,
            lb=problem.lb_init,
            ub=problem.ub_init,
            objective=problem.objective,
            x_guesses=problem.x_guesses,
            startpoint_method=startpoint_method,
        )

    return startpoints


def resample_startpoints(
    startpoints: np.ndarray,
    lb: np.ndarray,
    ub: np.ndarray,
    objective: ObjectiveBase,
    x_guesses: np.ndarray,
    startpoint_method: StartpointMethod,
):
    """Resample startpoints having non-finite value.

    Check all proposed startpoints and resample ones with non-finite value
    via the startpoint_method.

    Also order startpoints by function value, ascending.

    Parameters
    ----------
    startpoints:
        Previously proposed startpoints.
    lb:
        Lower parameter bound.
    ub:
        Upper parameter bound.
    objective:
        Objective function, required to evaluate function values.
    x_guesses:
        Externally provided guesses, may be needed to generate remote candidate
        points.
    startpoint_method:
        Startpoint generation method to use.

    Returns
    -------
    startpoints:
        Startpoints with all finite function values, shape (n_starts, n_par).
    """

    n_starts = startpoints.shape[0]
    resampled_startpoints = np.zeros_like(startpoints)

    fvals = np.empty((n_starts,))
    # iterate over startpoints
    for j in range(0, n_starts):
        startpoint = startpoints[j, :]
        # apply method until found valid point
        objective.initialize()
        fvals[j] = objective(startpoint)
        while fvals[j] == np.inf or fvals[j] == np.nan:
            startpoint = startpoint_method(
                n_starts=1,
                lb=lb,
                ub=ub,
                objective=objective,
            )[0, :]
            objective.initialize()
            fvals[j] = objective(startpoint)
        # assign startpoint
        resampled_startpoints[j, :] = startpoint

    # sort startpoints by function value, ascending
    startpoint_order = np.argsort(fvals)
    resampled_startpoints = resampled_startpoints[startpoint_order, :]

    return resampled_startpoints



import logging
from typing import Iterable, Union

from ..engine import Engine, SingleCoreEngine
from ..objective import HistoryOptions
from ..problem import Problem
from ..result import Result
from ..startpoint import (
    assign_startpoints, uniform, StartpointMethod, to_startpoint_method,
)
from .optimizer import Optimizer, ScipyOptimizer
from .options import OptimizeOptions
from .task import OptimizerTask
from .util import check_hdf5_mp, fill_hdf5_file

logger = logging.getLogger(__name__)


def minimize(
    problem: Problem,
    optimizer: Optimizer = None,
    n_starts: int = 100,
    ids: Iterable[str] = None,
    startpoint_method: Union[StartpointMethod, bool] = None,
    result: Result = None,
    engine: Engine = None,
    progress_bar: bool = True,
    options: OptimizeOptions = None,
    history_options: HistoryOptions = None,
) -> Result:
    """
    This is the main function to call to do multistart optimization.

    Parameters
    ----------
    problem:
        The problem to be solved.
    optimizer:
        The optimizer to be used n_starts times.
    n_starts:
        Number of starts of the optimizer.
    ids:
        Ids assigned to the startpoints.
    startpoint_method:
        Method for how to choose start points. False means the optimizer does
        not require start points, e.g. for the 'PyswarmOptimizer'.
    result:
        A result object to append the optimization results to. For example,
        one might append more runs to a previous optimization. If None,
        a new object is created.
    engine:
        Parallelization engine. Defaults to sequential execution on a
        SingleCoreEngine.
    progress_bar:
        Whether to display a progress bar.
    options:
        Various options applied to the multistart optimization.
    history_options:
        Optimizer history options.

    Returns
    -------
    result:
        Result object containing the results of all multistarts in
        `result.optimize_result`.
    """

    # optimizer
    if optimizer is None:
        optimizer = ScipyOptimizer()

    # startpoint method
    if startpoint_method is not None and problem.startpoint_method is not None:
        raise Warning(
            "Problem.startpoint_method will be ignored. Startpoints will be "
            "generated using the startpoint method given as an argument to "
            "the minimize function.",
        )
    elif problem.startpoint_method is not None:
        startpoint_method = problem.startpoint_method
    elif startpoint_method is None:
        startpoint_method = uniform

    # convert startpoint method to class instance
    startpoint_method = to_startpoint_method(startpoint_method)

    # check options
    if options is None:
        options = OptimizeOptions()
    options = OptimizeOptions.assert_instance(options)

    if history_options is None:
        history_options = HistoryOptions()
    history_options = HistoryOptions.assert_instance(history_options)

    # assign startpoints
    startpoints = assign_startpoints(
        n_starts=n_starts,
        startpoint_method=startpoint_method,
        problem=problem,
        startpoint_resample=options.startpoint_resample,
    )

    if ids is None:
        ids = [str(j) for j in range(n_starts)]
    if len(ids) != n_starts:
        raise AssertionError("Number of starts and ids must coincide.")

    # prepare result
    if result is None:
        result = Result(problem)

    # engine
    if engine is None:
        engine = SingleCoreEngine()

    # define tasks
    tasks = []
    filename = None
    if history_options.storage_file is not None and \
            history_options.storage_file.endswith(('.h5', '.hdf5')):
        filename = check_hdf5_mp(history_options, engine)

    for startpoint, id in zip(startpoints, ids):
        task = OptimizerTask(
            optimizer=optimizer,
            problem=problem,
            x0=startpoint,
            id=id,
            options=options,
            history_options=history_options,
        )
        tasks.append(task)

    # do multistart optimization
    ret = engine.execute(tasks, progress_bar=progress_bar)

    if filename is not None:
        fill_hdf5_file(ret, filename)

    # aggregate results
    for optimizer_result in ret:
        result.optimize_result.append(optimizer_result)

    # sort by best fval
    result.optimize_result.sort()

    return result



"""
Optimize
========

Multistart optimization with support for various optimizers.
"""

from .options import OptimizeOptions
from .optimize import (
    minimize)
from .optimizer import (
    Optimizer,
    ScipyOptimizer,
    IpoptOptimizer,
    DlibOptimizer,
    PyswarmOptimizer,
    CmaesOptimizer,
    PyswarmsOptimizer,
    ScipyDifferentialEvolutionOptimizer,
    NLoptOptimizer,
    FidesOptimizer,
    read_result_from_file,
)
from .result import OptimizerResult



"""
This is for (string) constants used in the objective module.
"""


MODE_FUN = 'mode_fun'  # mode for function values
MODE_RES = 'mode_res'  # mode for residuals
FVAL = 'fval'  # function value
FVAL0 = 'fval0'  # function value at start
GRAD = 'grad'  # gradient
HESS = 'hess'  # Hessian
HESSP = 'hessp'  # Hessian vector product
RES = 'res'  # residual
SRES = 'sres'  # residual sensitivities
RDATAS = 'rdatas'  # returned simulated data sets

TIME = 'time'  # time
N_FVAL = 'n_fval'  # number of function evaluations
N_GRAD = 'n_grad'  # number of gradient evaluations
N_HESS = 'n_hess'  # number of Hessian evaluations
N_RES = 'n_res'  # number of residual evaluations
N_SRES = 'n_sres'  # number of residual sensitivity evaluations
CHI2 = 'chi2'  # chi2 value
SCHI2 = 'schi2'  # chi2 value gradient
X = 'x'
X0 = 'x0'
ID = 'id'



import numpy as np
import logging

from ..engine import Task
from ..objective import HistoryOptions
from ..problem import Problem
import pypesto.optimize

logger = logging.getLogger(__name__)


class OptimizerTask(Task):
    """
    A multistart optimization task, performed in `pypesto.minimize`.
    """

    def __init__(
            self,
            optimizer: 'pypesto.optimize.Optimizer',
            problem: Problem,
            x0: np.ndarray,
            id: str,
            options: 'pypesto.optimize.OptimizeOptions',
            history_options: HistoryOptions):
        """
        Create the task object.

        Parameters
        ----------
        optimizer:
            The optimizer to use.
        problem:
            The problem to solve.
        x0:
            The point from which to start.
        id:
            The multistart id.
        options:
            Options object applying to optimization.
        history_options:
            Optimizer history options.
        """
        super().__init__()

        self.optimizer = optimizer
        self.problem = problem
        self.x0 = x0
        self.id = id
        self.options = options
        self.history_options = history_options

    def execute(self) -> 'pypesto.optimize.OptimizerResult':
        logger.info(f"Executing task {self.id}.")

        optimizer_result = self.optimizer.minimize(
            problem=self.problem, x0=self.x0, id=self.id,
            allow_failed_starts=self.options.allow_failed_starts,
            history_options=self.history_options)
        return optimizer_result



import numpy as np
import numbers
from typing import Dict, Sequence, Union, Tuple
import logging
import warnings

from .constants import (
    FVAL, CHI2, GRAD, HESS, RES, SRES, RDATAS, MODE_FUN, MODE_RES
)

try:
    import amici
    import amici.petab_objective
    import amici.parameter_mapping
    from amici.parameter_mapping import (
        ParameterMapping, ParameterMappingForCondition)
except ImportError:
    pass

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']

logger = logging.getLogger(__name__)


def map_par_opt_to_par_sim(
        condition_map_sim_var: Dict[str, Union[float, str]],
        x_dct: Dict[str, float],
        amici_model: AmiciModel
) -> np.ndarray:
    """
    From the optimization vector, create the simulation vector according
    to the mapping.

    Parameters
    ----------
    condition_map_sim_var:
        Simulation to optimization parameter mapping.
    x_dct:
        The optimization parameters dict.
    amici_model:
        The amici model.

    Returns
    -------
    par_sim_vals:
        The simulation parameters vector corresponding to x under the
        specified mapping.
    """
    par_sim_vals = [condition_map_sim_var[par_id]
                    for par_id in amici_model.getParameterIds()]

    # iterate over simulation parameter indices
    for ix, val in enumerate(par_sim_vals):
        if not isinstance(val, numbers.Number):
            # value is optimization parameter id
            par_sim_vals[ix] = x_dct[val]

    # return the created simulation parameter vector
    return np.array(par_sim_vals)


def create_plist_from_par_opt_to_par_sim(mapping_par_opt_to_par_sim):
    warnings.warn("This function will be removed in future releases. ",
                  DeprecationWarning)
    """
    From the parameter mapping `mapping_par_opt_to_par_sim`, create the
    simulation plist according to the mapping `mapping`.

    Parameters
    ----------

    mapping_par_opt_to_par_sim: array-like of str
        len == n_par_sim, the entries are either numeric, or
        optimization parameter ids.

    Returns
    -------
    plist: array-like of float
        List of parameter indices for which the sensitivity needs to be
        computed
    """
    plist = []

    # iterate over simulation parameter indices
    for j_par_sim, val in enumerate(mapping_par_opt_to_par_sim):
        if not isinstance(val, numbers.Number):
            plist.append(j_par_sim)

    # return the created simulation parameter vector
    return plist


def create_identity_parameter_mapping(
        amici_model: AmiciModel, n_conditions: int
) -> 'ParameterMapping':
    """Create a dummy identity parameter mapping table.

    This fills in only the dynamic parameters. Values for fixed parameters,
    both in preequilibration and simulation, are assumed to be provided
    correctly in model or edatas already.
    """
    x_ids = list(amici_model.getParameterIds())
    x_scales = list(amici_model.getParameterScale())
    parameter_mapping = ParameterMapping()
    for _ in range(n_conditions):
        condition_map_sim_var = {x_id: x_id for x_id in x_ids}
        condition_scale_map_sim_var = {
            x_id: amici.parameter_mapping.amici_to_petab_scale(x_scale)
            for x_id, x_scale in zip(x_ids, x_scales)}
        # assumes fixed parameters are filled in already
        mapping_for_condition = ParameterMappingForCondition(
            map_sim_var=condition_map_sim_var,
            scale_map_sim_var=condition_scale_map_sim_var)

        parameter_mapping.append(mapping_for_condition)
    return parameter_mapping


def par_index_slices(
        par_opt_ids: Sequence[str], par_sim_ids: Sequence[str],
        condition_map_sim_var: Dict[str, Union[float, str]]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate numpy arrays for indexing based on `mapping_par_opt_to_par_sim`.

    Parameters
    ----------
    par_opt_ids:
        The optimization parameter ids. Needed for order.
    par_sim_ids:
        The simulation parameter ids. Needed for order.
    condition_map_sim_var:
        The simulation to optimization parameter mapping.

    Returns
    ----------
    par_sim_slice:
        array of simulation parameter indices

    par_opt_slice:
        array of optimization parameter indices
    """
    # the sum accounts for subindexing according to plist in edata
    par_sim_slice, par_opt_slice = list(zip(*[
        (sum(isinstance(condition_map_sim_var[par_id], str)
             for par_id in
             par_sim_ids[:par_sim_ids.index(par_sim_id) + 1]) - 1,
         par_opt_ids.index(par_opt_id))
        for par_sim_id, par_opt_id in condition_map_sim_var.items()
        if isinstance(par_opt_id, str)
    ]))
    return np.asarray(par_sim_slice), np.asarray(par_opt_slice)


def add_sim_grad_to_opt_grad(
        par_opt_ids: Sequence[str],
        par_sim_ids: Sequence[str],
        condition_map_sim_var: Dict[str, Union[float, str]],
        sim_grad: np.ndarray,
        opt_grad: np.ndarray,
        coefficient: float = 1.0):
    """
    Sum simulation gradients to objective gradient according to the provided
    mapping `mapping_par_opt_to_par_sim`.

    Parameters
    ----------
    par_opt_ids:
        The optimization parameter ids. Needed for order.
    par_sim_ids:
        The simulation parameter ids. Needed for order.
    condition_map_sim_var:
        The simulation to optimization parameter mapping.
    sim_grad:
        Simulation gradient.
    opt_grad:
        The optimization gradient. To which sim_grad is added.
        Changed in-place.
    coefficient:
        Coefficient for sim_grad when adding to opt_grad.
    """

    par_sim_slice, par_opt_slice = par_index_slices(par_opt_ids, par_sim_ids,
                                                    condition_map_sim_var)

    par_opt_slice_unique, unique_index = np.unique(par_opt_slice,
                                                   return_index=True)
    opt_grad[par_opt_slice_unique] += \
        coefficient * sim_grad[par_sim_slice[unique_index]]

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in range(len(par_opt_slice)):
            if idx not in unique_index:
                opt_grad[par_opt_slice[idx]] += \
                    coefficient * sim_grad[par_sim_slice[idx]]


def add_sim_hess_to_opt_hess(
        par_opt_ids: Sequence[str],
        par_sim_ids: Sequence[str],
        condition_map_sim_var: Dict[str, Union[float, str]],
        sim_hess: np.ndarray,
        opt_hess: np.ndarray,
        coefficient: float = 1.0):
    """
    Sum simulation hessians to objective hessian according to the provided
    mapping `mapping_par_opt_to_par_sim`.

    Parameters
    ----------
    Same as for add_sim_grad_to_opt_grad, replacing the gradients by hessians.
    """

    par_sim_slice, par_opt_slice = par_index_slices(par_opt_ids, par_sim_ids,
                                                    condition_map_sim_var)

    par_opt_slice_unique, unique_index = np.unique(par_opt_slice,
                                                   return_index=True)

    non_unique_indices = [idx for idx in range(len(par_opt_slice))
                          if idx not in unique_index]

    opt_hess[np.ix_(par_opt_slice_unique, par_opt_slice_unique)] += \
        coefficient * sim_hess[np.ix_(par_sim_slice[unique_index],
                                      par_sim_slice[unique_index])]

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in non_unique_indices:
            opt_hess[par_opt_slice[idx], par_opt_slice_unique] += \
                coefficient * sim_hess[par_sim_slice[idx],
                                       par_sim_slice[unique_index]]
            opt_hess[par_opt_slice_unique, par_opt_slice[idx]] += \
                coefficient * sim_hess[par_sim_slice[unique_index],
                                       par_sim_slice[idx]]
            for jdx in non_unique_indices:
                opt_hess[par_opt_slice[idx], par_opt_slice[jdx]] += \
                    coefficient * sim_hess[par_sim_slice[idx],
                                           par_sim_slice[jdx]]


def sim_sres_to_opt_sres(par_opt_ids: Sequence[str],
                         par_sim_ids: Sequence[str],
                         condition_map_sim_var: Dict[str, Union[float, str]],
                         sim_sres: np.ndarray,
                         coefficient: float = 1.0):
    """
    Sum simulation residual sensitivities to objective residual sensitivities
    according to the provided mapping.

    Parameters
    ----------
    Mostly the same as for add_sim_grad_to_opt_grad, replacing the gradients by
    residual sensitivities.
    """
    opt_sres = np.zeros((sim_sres.shape[0], len(par_opt_ids)))

    par_sim_slice, par_opt_slice = par_index_slices(par_opt_ids, par_sim_ids,
                                                    condition_map_sim_var)

    par_opt_slice_unique, unique_index = np.unique(par_opt_slice,
                                                   return_index=True)
    opt_sres[:, par_opt_slice_unique] += \
        coefficient * sim_sres[:, par_sim_slice[unique_index]]

    if par_opt_slice_unique.size < par_opt_slice.size:
        for idx in range(len(par_opt_slice)):
            if idx not in unique_index:
                opt_sres[:, par_opt_slice[idx]] += \
                    coefficient * sim_sres[:, par_sim_slice[idx]]

    return opt_sres


def log_simulation(data_ix, rdata):
    """Log the simulation results."""
    logger.debug(f"=== DATASET {data_ix} ===")
    logger.debug(f"status: {rdata['status']}")
    logger.debug(f"llh: {rdata['llh']}")

    t_steadystate = 't_steadystate'
    if t_steadystate in rdata and rdata[t_steadystate] != np.nan:
        logger.debug(f"t_steadystate: {rdata[t_steadystate]}")

    logger.debug(f"res: {rdata['res']}")


def get_error_output(
        amici_model: AmiciModel,
        edatas: Sequence['amici.ExpData'],
        rdatas: Sequence['amici.ReturnData'],
        sensi_order: int,
        mode: str,
        dim: int):
    """Default output upon error.

    Returns values indicative of an error, that is with nan entries in all
    vectors, and a function value, i.e. nllh, of `np.inf`.
    """
    if not amici_model.nt():
        nt = sum(data.nt() for data in edatas)
    else:
        nt = sum(data.nt() if data.nt() else amici_model.nt()
                 for data in edatas)
    n_res = nt * amici_model.nytrue

    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(sensi_order,
                                                              mode, dim, True)
    if res is not None:
        res = np.nan * np.ones(n_res)
    if sres is not None:
        sres = np.nan * np.ones([n_res, dim])

    ret = {
        FVAL: nllh,
        CHI2: chi2,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas
    }
    return filter_return_dict(ret)


def init_return_values(sensi_order, mode, dim, error=False):
    if error:
        fval = np.inf
        sval = np.nan
    else:
        fval = sval = 0.0

    nllh = fval
    snllh = None
    s2nllh = None
    if mode == MODE_FUN and sensi_order > 0:
        snllh = sval * np.ones(dim)
        if sensi_order > 1:
            s2nllh = sval * np.ones([dim, dim])

    chi2 = None
    res = None
    sres = None
    if mode == MODE_RES:
        chi2 = fval
        res = np.zeros([0])
        if sensi_order > 0:
            sres = np.zeros([0, dim])

    return nllh, snllh, s2nllh, chi2, res, sres


def filter_return_dict(ret):
    """Filters return dict for non-None values"""
    return {
        key: val
        for key, val in ret.items()
        if val is not None
    }



import numpy as np
from typing import Dict, List, Sequence, Union

from .constants import (
    MODE_FUN, MODE_RES, FVAL, GRAD, HESS, RES, SRES, RDATAS, CHI2
)
from .amici_util import (
    add_sim_grad_to_opt_grad, add_sim_hess_to_opt_hess,
    sim_sres_to_opt_sres, log_simulation, get_error_output, filter_return_dict,
    init_return_values,
)

try:
    import amici
    import amici.petab_objective
    import amici.parameter_mapping
    from amici.parameter_mapping import ParameterMapping
except ImportError:
    pass

AmiciModel = Union['amici.Model', 'amici.ModelPtr']
AmiciSolver = Union['amici.Solver', 'amici.SolverPtr']


class AmiciCalculator:
    """
    Class to perform the actual call to AMICI and obtain requested objective
    function values.
    """

    def __init__(self):
        self._known_least_squares_safe = False

    def initialize(self):
        """Initialize the calculator. Default: Do nothing."""

    def __call__(self,
                 x_dct: Dict,
                 sensi_order: int,
                 mode: str,
                 amici_model: AmiciModel,
                 amici_solver: AmiciSolver,
                 edatas: List['amici.ExpData'],
                 n_threads: int,
                 x_ids: Sequence[str],
                 parameter_mapping: 'ParameterMapping',
                 fim_for_hess: bool):
        """Perform the actual AMICI call.

        Called within the :func:`AmiciObjective.__call__` method.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_order:
            Maximum sensitivity order.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        """
        # set order in solver
        if sensi_order == 2 and fim_for_hess:
            # we use the FIM
            amici_solver.setSensitivityOrder(sensi_order-1)
        else:
            amici_solver.setSensitivityOrder(sensi_order)

        # fill in parameters
        amici.parameter_mapping.fill_in_parameters(
            edatas=edatas,
            problem_parameters=x_dct,
            scaled_parameters=True,
            parameter_mapping=parameter_mapping,
            amici_model=amici_model
        )

        # run amici simulation
        rdatas = amici.runAmiciSimulations(
            amici_model,
            amici_solver,
            edatas,
            num_threads=min(n_threads, len(edatas)),
        )
        if not self._known_least_squares_safe and mode == MODE_RES and \
                sensi_order > 0:
            if not amici_model.getAddSigmaResiduals() and any(
                ((r['ssigmay'] is not None and np.any(r['ssigmay']))
                 or
                 (r['ssigmaz'] is not None and np.any(r['ssigmaz'])))
                for r in rdatas
            ):
                raise RuntimeError('Cannot use least squares solver with'
                                   'parameter dependent sigma! Support can be '
                                   'enabled via '
                                   'amici_model.setAddSigmaResiduals().')
            self._known_least_squares_safe = True  # don't check this again

        return calculate_function_values(
            rdatas=rdatas, sensi_order=sensi_order, mode=mode,
            amici_model=amici_model, amici_solver=amici_solver, edatas=edatas,
            x_ids=x_ids, parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess)


def calculate_function_values(rdatas,
                              sensi_order: int,
                              mode: str,
                              amici_model: AmiciModel,
                              amici_solver: AmiciSolver,
                              edatas: List['amici.ExpData'],
                              x_ids: Sequence[str],
                              parameter_mapping: 'ParameterMapping',
                              fim_for_hess: bool):
    # full optimization problem dimension (including fixed parameters)
    dim = len(x_ids)

    # check if the simulation failed
    if any(rdata['status'] < 0.0 for rdata in rdatas):
        return get_error_output(amici_model, edatas, rdatas,
                                sensi_order, mode, dim)

    nllh, snllh, s2nllh, chi2, res, sres = init_return_values(sensi_order,
                                                              mode, dim)

    par_sim_ids = list(amici_model.getParameterIds())
    sensi_method = amici_solver.getSensitivityMethod()

    # iterate over return data
    for data_ix, rdata in enumerate(rdatas):
        log_simulation(data_ix, rdata)

        condition_map_sim_var = \
            parameter_mapping[data_ix].map_sim_var

        # add objective value
        nllh -= rdata['llh']

        if mode == MODE_FUN:
            if not np.isfinite(nllh):
                return get_error_output(amici_model, edatas, rdatas,
                                        sensi_order, mode, dim)

            if sensi_order > 0:
                # add gradient
                add_sim_grad_to_opt_grad(
                    x_ids,
                    par_sim_ids,
                    condition_map_sim_var,
                    rdata['sllh'],
                    snllh,
                    coefficient=-1.0
                )

                if not np.isfinite(snllh).all():
                    return get_error_output(amici_model, edatas, rdatas,
                                            sensi_order, mode, dim)

                # Hessian
                if sensi_order > 1:
                    if sensi_method == amici.SensitivityMethod_forward \
                            and fim_for_hess:
                        # add FIM for Hessian
                        add_sim_hess_to_opt_hess(
                            x_ids,
                            par_sim_ids,
                            condition_map_sim_var,
                            rdata['FIM'],
                            s2nllh,
                            coefficient=+1.0
                        )
                        if not np.isfinite(s2nllh).all():
                            return get_error_output(amici_model, edatas,
                                                    rdatas, sensi_order,
                                                    mode, dim)
                    else:
                        raise ValueError("AMICI cannot compute Hessians yet.")

        elif mode == MODE_RES:
            chi2 += rdata['chi2']
            res = np.hstack([res, rdata['res']]) \
                if res.size else rdata['res']
            if sensi_order > 0:
                opt_sres = sim_sres_to_opt_sres(
                    x_ids,
                    par_sim_ids,
                    condition_map_sim_var,
                    rdata['sres'],
                    coefficient=1.0
                )
                sres = np.vstack([sres, opt_sres]) \
                    if sres.size else opt_sres

    ret = {
        FVAL: nllh,
        CHI2: chi2,
        GRAD: snllh,
        HESS: s2nllh,
        RES: res,
        SRES: sres,
        RDATAS: rdatas
    }
    return filter_return_dict(ret)



import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
import numpy as np
import pandas as pd

from typing import Optional, Tuple

from ..ensemble import Ensemble
from ..ensemble.constants import (
    COLOR_HIT_BOTH_BOUNDS, COLOR_HIT_ONE_BOUND, COLOR_HIT_NO_BOUNDS)


def ensemble_identifiability(ensemble: Ensemble,
                             ax: Optional[plt.Axes] = None,
                             size: Optional[Tuple[float]] = (12, 6)):
    """
    Plots an overview about how many parameters hit the parameter bounds based
    on a ensemble of parameters. confidence intervals/credible ranges are
    computed via the ensemble mean plus/minus 1 standard deviation.
    This highlevel routine expects a ensemble object as input.

    Parameters
    ----------

    ensemble:
        ensemble of parameter vectors (from pypesto.ensemble)

    ax:
        Axes object to use.

    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # first get the data to check identifiability
    id_df = ensemble.check_identifiability()

    # check how many bounds are actually hit and which ones
    none_hit, lb_hit, ub_hit, both_hit = _prepare_identifiability_plot(id_df)

    # call lowlevel routine whick works with np arrays only
    ax = ensemble_identifiability_lowlevel(none_hit, lb_hit, ub_hit, both_hit,
                                           ax, size)

    return ax


def ensemble_identifiability_lowlevel(none_hit: np.ndarray,
                                      lb_hit: np.ndarray,
                                      ub_hit: np.ndarray,
                                      both_hit: np.ndarray,
                                      ax: Optional[plt.Axes] = None,
                                      size: Optional[Tuple[float]] = (16, 10)):
    """
    Plots an overview about how many parameters hit the parameter bounds based
    on a ensemble of parameters. Confidence intervals/credible ranges are
    computed via the ensemble mean plus/minus 1 standard deviation.
    This lowlevel routine works with numpy arrays which define the confidence
    intervals/credible ranges of each parameter.

    Parameters
    ----------

    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters

    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound

    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound

    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds

    ax:
        Axes object to use.

    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # define some short hands for later plotting
    n_par = sum([none_hit.shape[0], lb_hit.shape[0],
                 ub_hit.shape[0], both_hit.shape[0]])
    x_both = len(both_hit) / n_par
    x_lb = len(lb_hit) / n_par
    x_ub = len(ub_hit) / n_par
    x_none = 1. - x_both - x_ub - x_lb

    patches_both_hit, patches_lb_hit, patches_ub_hit, patches_none_hit = \
        _create_patches(none_hit, lb_hit, ub_hit, both_hit)

    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # create axes object and add patch collections
    if patches_both_hit:
        ax.add_collection(patches_both_hit)
    if patches_lb_hit:
        ax.add_collection(patches_lb_hit)
    if patches_ub_hit:
        ax.add_collection(patches_ub_hit)
    if patches_none_hit:
        ax.add_collection(patches_none_hit)

    # plot dashed lines indicating the number rof non-identifiable parameters
    vert = [-.05, 1.05]
    ax.plot([x_both, x_both], vert, 'k--', linewidth=1.5)
    ax.plot([x_both + x_lb, x_both + x_lb], vert, 'k--', linewidth=1.5)
    ax.plot([x_both + x_lb + x_ub, x_both + x_lb + x_ub], vert,
            'k--', linewidth=1.5)

    # add text
    if patches_both_hit:
        ax.text(x_both / 2, -.05, 'both bounds hit',
                color=COLOR_HIT_BOTH_BOUNDS,
                rotation=-90, va='top', ha='center')
    if patches_lb_hit:
        ax.text(x_both + x_lb / 2, -.05, 'lower bound hit',
                color=COLOR_HIT_ONE_BOUND, rotation=-90, va='top', ha='center')
    if patches_ub_hit:
        ax.text(x_both + x_lb + x_ub / 2, -.05, 'upper bound hit',
                color=COLOR_HIT_ONE_BOUND, rotation=-90, va='top', ha='center')
    if patches_none_hit:
        ax.text(1 - x_none / 2, -.05, 'no bounds hit',
                color=COLOR_HIT_NO_BOUNDS, rotation=-90, va='top', ha='center')
    ax.text(0, -.7, 'identifiable parameters: {:4.1f}%'.format(x_none * 100),
            va='top')

    # plot upper and lower bounds
    ax.text(-.03, 1., 'upper\nbound', ha='right', va='center')
    ax.text(-.03, 0., 'lower\nbound', ha='right', va='center')
    ax.plot([-.02, 1.03], [0, 0], 'k:', linewidth=1.5)
    ax.plot([-.02, 1.03], [1, 1], 'k:', linewidth=1.5)
    plt.xticks([])
    plt.yticks([])

    # plot frame
    ax.plot([0, 0], vert, 'k-', linewidth=1.5)
    ax.plot([1, 1], vert, 'k-', linewidth=1.5)

    # beautify axes
    plt.xlim((-.15, 1.1))
    plt.ylim((-.78, 1.15))
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)

    return ax


def _prepare_identifiability_plot(id_df: pd.DataFrame):
    """
    This routine groups model parameters based on a ensemble object into
    four categories, based on the mean of the parameter ensemble plus/minus
    1 standard deviation: Parameters that hit both bounds, parameters that hit
    only the lower [or upper] bound, and parameters that hit no bounds.
    It returns them as four numpy arrays, together with their confidence
    intervals/credible ranges.

    Parameters
    ----------
    id_df:
        Pandas dataframe with information about parameter identifiability,
        as created by pypesto.ensemble.check_identifiability()

    Returns
    -------
    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters

    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound

    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound

    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds
    """

    # prepare
    both_hit = []
    lb_hit = []
    ub_hit = []
    none_hit = []

    def _affine_transform(par_info):
        # rescale parameters to bounds
        lb = par_info['lowerBound']
        ub = par_info['upperBound']
        val_l = par_info['ensemble_mean'] - par_info['ensemble_std']
        val_u = par_info['ensemble_mean'] + par_info['ensemble_std']
        # check if parameter confidence intervals/credible ranges hit bound
        if val_l <= lb:
            lower_val = 0.
        else:
            lower_val = (val_l - lb) / (ub - lb)
        if val_u >= ub:
            upper_val = 1.
        else:
            upper_val = (val_u - lb) / (ub - lb)

        return lower_val, upper_val

    for par_id in list(id_df.index):
        # check which of the parameters seems to be identifiable and group them
        if id_df.loc[par_id, 'within lb: 1 std'] and \
                id_df.loc[par_id, 'within ub: 1 std']:
            none_hit.append(_affine_transform(id_df.loc[par_id, :]))
        elif id_df.loc[par_id, 'within lb: 1 std']:
            ub_hit.append(_affine_transform(id_df.loc[par_id, :]))
        elif id_df.loc[par_id, 'within ub: 1 std']:
            lb_hit.append(_affine_transform(id_df.loc[par_id, :]))
        else:
            both_hit.append(_affine_transform(id_df.loc[par_id, :]))

    return np.array(none_hit), np.array(lb_hit), np.array(ub_hit), \
        np.array(both_hit)


def _create_patches(none_hit: np.ndarray,
                    lb_hit: np.ndarray,
                    ub_hit: np.ndarray,
                    both_hit: np.ndarray):
    """
    Creates matplotlib.patches.PatchCollection objects from numpy arrays with
    confidence intervals/credible ranges, which visualize identifiability
    properties of the model parameters.

    Parameters
    ----------
    none_hit:
        2-dimensional array of confidence interval/credible ranges for
        identifiable parameters

    lb_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the lower parameter bound

    ub_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit the upper parameter bound

    both_hit:
        2-dimensional array of confidence interval/credible ranges for
        parameters which hit both parameter bounds

    Returns
    -------
    patches_both_hit:
        patches showing parameters which hit both parameter bounds in the
        ensemble (and are hence non-identifiable)

    patches_lb_hit:
        patches showing parameters which hit only the lower parameter bounds
        in the ensemble (and are hence non-identifiable)

    patches_ub_hit:
        patches showing parameters which hit only the lower parameter bounds
        in the ensemble (and are hence non-identifiable)

    patches_none_hit
        patches showing parameters which hit no parameter bounds in the
        ensemble (and are hence identifiable)
    """
    # get total number of parameters
    n_par = sum([none_hit.shape[0], lb_hit.shape[0],
                 ub_hit.shape[0], both_hit.shape[0]])

    # start patches at the left end and increment by h = 1/n_par
    x = 0.
    h = 1. / n_par

    # creates patches for parameters which hit both bounds
    patches_both_hit = []
    if both_hit.size > 0:
        for _ in both_hit:
            # create a list of rectangles
            patches_both_hit.append(Rectangle((x, 0.), h, 1.))
            x += h
        patches_both_hit = PatchCollection(patches_both_hit,
                                           facecolors=COLOR_HIT_BOTH_BOUNDS)

    # creates patches for parameters which hit lower bound
    patches_lb_hit = []
    # sort by normalizes length of confidence interval/credible range
    if lb_hit.size > 0:
        tmp_lb = np.sort(lb_hit[:, 1])[::-1]
        for lb_par in tmp_lb:
            # create a list of rectangles
            patches_lb_hit.append(Rectangle((x, 0.), h, lb_par))
            x += h
        patches_lb_hit = PatchCollection(patches_lb_hit,
                                         facecolors=COLOR_HIT_ONE_BOUND)

    # creates patches for parameters which hit upper bound
    patches_ub_hit = []
    # sort by normalizes length of confidence interval/credible range
    if ub_hit.size > 0:
        tmp_ub = np.sort(ub_hit[:, 0])
        for ub_par in tmp_ub:
            # create a list of rectangles
            patches_ub_hit.append(Rectangle((x, ub_par), h, 1. - ub_par))
            x += h
        patches_ub_hit = PatchCollection(patches_ub_hit,
                                         facecolors=COLOR_HIT_ONE_BOUND)

    # creates patches for parameters which hit no bounds
    patches_none_hit = []
    # sort by normalizes length of confidence interval/credible range
    if none_hit.size > 0:
        tmp_none = np.argsort(none_hit[:, 1] - none_hit[:, 0])[::-1]
        for none_par in tmp_none:
            patches_none_hit.append(
                # create a list of rectangles
                Rectangle((x, none_hit[none_par, 0]), h,
                          none_hit[none_par, 1] - none_hit[none_par, 0]))
            x += h
        patches_none_hit = PatchCollection(patches_none_hit,
                                           facecolors=COLOR_HIT_NO_BOUNDS)

    return patches_both_hit, patches_lb_hit, patches_ub_hit, patches_none_hit



import logging
from typing import Callable, Union, Iterable

from ..engine import Engine, SingleCoreEngine
from ..optimize import Optimizer
from ..problem import Problem
from ..result import Result
from .profile_next_guess import next_guess
from .options import ProfileOptions
from .util import initialize_profile
from .task import ProfilerTask

logger = logging.getLogger(__name__)


def parameter_profile(
        problem: Problem,
        result: Result,
        optimizer: Optimizer,
        engine: Engine = None,
        profile_index: Iterable[int] = None,
        profile_list: int = None,
        result_index: int = 0,
        next_guess_method: Union[Callable, str] = 'adaptive_step_regression',
        profile_options: ProfileOptions = None,
        progress_bar: bool = True
) -> Result:
    """
    This is the main function to call to do parameter profiling.

    Parameters
    ----------
    problem:
        The problem to be solved.
    result:
        A result object to initialize profiling and to append the profiling
        results to. For example, one might append more profiling runs to a
        previous profile, in order to merge these.
        The existence of an optimization result is obligatory.
    optimizer:
        The optimizer to be used along each profile.
    engine:
        The engine to be used.
    profile_index:
        List with the parameter indices to be profiled
        (by default all free indices).
    profile_list:
        Integer which specifies whether a call to the profiler should create
        a new list of profiles (default) or should be added to a specific
        profile list.
    result_index:
        Index from which optimization result profiling should be started
        (default: global optimum, i.e., index = 0).
    next_guess_method:
        Function handle to a method that creates the next starting point for
        optimization in profiling.
    profile_options:
        Various options applied to the profile optimization.
    progress_bar:
        Whether to display a progress bar.

    Returns
    -------
    result:
        The profile results are filled into `result.profile_result`.
    """
    # Handling defaults
    # profiling indices
    if profile_index is None:
        profile_index = problem.x_free_indices

    # check profiling options
    if profile_options is None:
        profile_options = ProfileOptions()
    profile_options = ProfileOptions.create_instance(profile_options)

    # create a function handle that will be called later to get the next point
    if isinstance(next_guess_method, str):
        def create_next_guess(x, par_index, par_direction_, profile_options_,
                              current_profile_, problem_, global_opt_):
            return next_guess(x, par_index, par_direction_, profile_options_,
                              next_guess_method, current_profile_, problem_,
                              global_opt_)
    elif callable(next_guess_method):
        raise Exception('Passing function handles for computation of next '
                        'profiling point is not yet supported.')
    else:
        raise Exception('Unsupported input for next_guess_method.')

    # create the profile result object (retrieve global optimum) or append to
    # existing list of profiles
    global_opt = initialize_profile(problem, result, result_index,
                                    profile_index, profile_list)
    # if engine==None set SingleCoreEngine() as default
    if engine is None:
        engine = SingleCoreEngine()

    # create Tasks
    tasks = []
    # loop over parameters to create tasks
    for i_par in profile_index:
        # only compute profiles for free parameters
        if i_par in problem.x_fixed_indices:
            continue

        current_profile = result.profile_result.get_profiler_result(
            i_par=i_par, profile_list=profile_list)

        task = ProfilerTask(
            current_profile=current_profile,
            problem=problem,
            optimizer=optimizer,
            options=profile_options,
            create_next_guess=create_next_guess,
            global_opt=global_opt,
            i_par=i_par
        )
        tasks.append(task)

    # execute the tasks with Engine
    indexed_profiles = engine.execute(tasks, progress_bar=progress_bar)

    # fill in the ProfilerResults at the right index
    for indexed_profile in indexed_profiles:
        result.profile_result.list[-1][indexed_profile['index']] = \
            indexed_profile['profile']

    return result



import os
import logging
from typing import Union
from numbers import Integral

import h5py
import numpy as np

from .hdf5 import (write_array, write_float_array)
from ..result import Result, SampleResult

logger = logging.getLogger(__name__)


def check_overwrite(f: Union[h5py.File, h5py.Group],
                    overwrite: bool,
                    target: str):
    """
    Checks whether target already exists. Sends a warning if
    overwrite=False, deletes the target if overwrite=True.

    Attributes
    -------------
    f: file or group where existence of a group with the path group_path
       should be checked
    target: name of the group, whose existence is checked
    overwrite: if overwrite is true, it deltes the target in f
    """
    if target in f:
        if overwrite:
            del f[target]
        else:
            raise Exception(f"The file already exists and contains "
                            f"information about {target} result."
                            f"If you wish to overwrite the file set"
                            f"overwrite=True.")


class ProblemHDF5Writer:
    """
    Writer of the HDF5 problem files.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 problem file name
        """
        self.storage_filename = storage_filename

    def write(self, problem, overwrite: bool = False):
        """
        Write HDF5 problem file from pyPESTO problem object.
        """

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'problem')
            attrs_to_save = [a for a in dir(problem) if not a.startswith('__')
                             and not callable(getattr(problem, a))
                             and not hasattr(type(problem), a)]

            problem_grp = f.create_group("problem")
            # save the configuration
            f['problem/config'] = str(problem.objective.get_config())

            for problem_attr in attrs_to_save:
                value = getattr(problem, problem_attr)
                if isinstance(value, (list, np.ndarray)):
                    value = np.asarray(value)
                    if value.size:
                        write_array(problem_grp, problem_attr, value)
                elif isinstance(value, Integral):
                    problem_grp.attrs[problem_attr] = value


def get_or_create_group(f: Union[h5py.File, h5py.Group],
                        group_path: str) -> h5py.Group:
    """
    Helper function that returns a group object for the group with group_path
    relative to f. Creates it if it doesn't exist.

    Attributes
    -------------
    f: file or group where existence of a group with the path group_path
       should be checked
    group_path: the path or simply the name of the group that should exist in f

    Returns
    -------
    grp:
        hdf5 group object with specified path.
    """
    if group_path in f:
        grp = f[group_path]
    else:
        grp = f.create_group(group_path)
    return grp


class OptimizationResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite=False):
        """
        Write HDF5 result file from pyPESTO result object.
        """

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'optimization')
            optimization_grp = get_or_create_group(f, "optimization")
            # settings =
            # optimization_grp.create_dataset("settings", settings, dtype=)
            results_grp = get_or_create_group(optimization_grp, "results")

            for start in result.optimize_result.list:
                start_id = start['id']
                start_grp = get_or_create_group(results_grp, start_id)
                for key in start.keys():
                    if key == 'history':
                        continue
                    if isinstance(start[key], np.ndarray):
                        write_float_array(start_grp, key, start[key])
                    elif start[key] is not None:
                        start_grp.attrs[key] = start[key]
                f.flush()


class SamplingResultHDF5Writer:
    """
    Writer of the HDF5 sampling files.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite: bool = False):
        """
        Write HDF5 sampling file from pyPESTO result object.
        """
        # if there is no sample available, log a warning and return
        # SampleResult is only a dummy class created by the Result class
        # and always indicates the lack of a sampling result.
        if(isinstance(result.sample_result, SampleResult)):
            logger.warning("Warning: There is no sampling_result, "
                           "which you tried to save to hdf5.")
            return

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'sampling')
            sampling_grp = get_or_create_group(f, "sampling")
            results_grp = get_or_create_group(sampling_grp, "results")

            for key in result.sample_result.keys():
                if isinstance(result.sample_result[key], np.ndarray):
                    write_float_array(results_grp,
                                      key,
                                      result.sample_result[key])
                elif result.sample_result[key] is not None:
                    results_grp.attrs[key] = result.sample_result[key]
            f.flush()


class ProfileResultHDF5Writer:
    """
    Writer of the HDF5 result files.

    Attributes
    -------------
    storage_filename:
        HDF5 result file name
    """

    def __init__(self, storage_filename: str):
        """
        Parameters
        ----------

        storage_filename: str
            HDF5 result file name
        """
        self.storage_filename = storage_filename

    def write(self, result: Result, overwrite: bool = False):
        """
        Write HDF5 result file from pyPESTO result object.
        """

        # Create destination directory
        if isinstance(self.storage_filename, str):
            basedir = os.path.dirname(self.storage_filename)
            if basedir:
                os.makedirs(basedir, exist_ok=True)

        with h5py.File(self.storage_filename, "a") as f:
            check_overwrite(f, overwrite, 'profiling')
            profiling_grp = get_or_create_group(f, "profiling")

            for profile_id, profile in enumerate(result.profile_result.list):
                profile_grp = get_or_create_group(profiling_grp,
                                                  str(profile_id))
                for parameter_id, parameter_profile in enumerate(profile):
                    result_grp = get_or_create_group(profile_grp,
                                                     str(parameter_id))

                    if parameter_profile is None:
                        result_grp.attrs['IsNone'] = True
                        continue
                    result_grp.attrs['IsNone'] = False
                    for key in parameter_profile.keys():
                        if isinstance(parameter_profile[key], np.ndarray):
                            write_float_array(result_grp,
                                              key,
                                              parameter_profile[key])
                        elif parameter_profile[key] is not None:
                            result_grp.attrs[key] = parameter_profile[key]
            f.flush()


def write_result(result: Result,
                 filename: str,
                 overwrite: bool = False,
                 problem: bool = True,
                 optimize: bool = True,
                 profile: bool = True,
                 sample: bool = True,
                 ):
    """Save whole pypesto.Result to hdf5 file.

    Boolean indicators allow specifying what to save.

    Parameters
    ----------
    result:
        The :class:`pypesto.Result` object to be saved.
    filename:
        The HDF5 filename.
    overwrite:
        Boolean, whether already existing results should be overwritten.
    problem:
        Read the problem.
    optimize:
        Read the optimize result.
    profile:
        Read the profile result.
    sample:
        Read the sample result.
    """

    if problem:
        pypesto_problem_writer = ProblemHDF5Writer(filename)
        pypesto_problem_writer.write(result.problem, overwrite=overwrite)

    if optimize:
        pypesto_opt_writer = OptimizationResultHDF5Writer(filename)
        pypesto_opt_writer.write(result, overwrite=overwrite)

    if profile:
        pypesto_profile_writer = ProfileResultHDF5Writer(filename)
        pypesto_profile_writer.write(result, overwrite=overwrite)

    if sample:
        pypesto_sample_writer = SamplingResultHDF5Writer(filename)
        pypesto_sample_writer.write(result, overwrite=overwrite)



"""
Problem
=======

A problem contains the objective as well as all information like prior
describing the problem to be solved.

"""

import numpy as np
import pandas as pd
import copy

from typing import Iterable, List, Optional, Union, SupportsFloat, \
    SupportsInt, Callable

from .objective import ObjectiveBase
from .objective.priors import NegLogPriors

SupportsFloatIterableOrValue = Union[Iterable[SupportsFloat], SupportsFloat]
SupportsIntIterableOrValue = Union[Iterable[SupportsInt], SupportsInt]


class Problem:
    """
    The problem formulation. A problem specifies the objective function,
    boundaries and constraints, parameter guesses as well as the parameters
    which are to be optimized.

    Parameters
    ----------
    objective:
        The objective function for minimization. Note that a shallow copy
        is created.
    lb, ub:
        The lower and upper bounds for optimization. For unbounded directions
        set to +-inf.
    lb_init, ub_init:
        The lower and upper bounds for initialization, typically for defining
        search start points.
        If not set, set to lb, ub.
    dim_full:
        The full dimension of the problem, including fixed parameters.
    x_fixed_indices:
        Vector containing the indices (zero-based) of parameter components
        that are not to be optimized.
    x_fixed_vals:
        Vector of the same length as x_fixed_indices, containing the values
        of the fixed parameters.
    x_guesses:
        Guesses for the parameter values, shape (g, dim), where g denotes the
        number of guesses. These are used as start points in the optimization.
    startpoint_method:
        Callable. `startpoint_method(n_starts)` returns a
        n_starts x n_free_indices array of initial values for the optimization.
    x_names:
        Parameter names that can be optionally used e.g. in visualizations.
        If objective.get_x_names() is not None, those values are used,
        else the values specified here are used if not None, otherwise
        the variable names are set to ['x0', ... 'x{dim_full}']. The list
        must always be of length dim_full.
    x_scales:
        Parameter scales can be optionally given and are used e.g. in
        visualisation and prior generation. Currently the scales 'lin',
        'log`and 'log10' are supported.
    x_priors_defs:
        Definitions of priors for parameters. Types of priors, and their
        required and optional parameters, are described in the `Prior` class.
    copy_objective:
        Whethter to generate a deep copy of the objective function before
        potential modification the problem class performs on it.

    Notes
    -----

    On the fixing of parameter values:

    The number of parameters dim_full the objective takes as input must
    be known, so it must be either lb a vector of that size, or dim_full
    specified as a parameter.

    All vectors are mapped to the reduced space of dimension dim in __init__,
    regardless of whether they were in dimension dim or dim_full before. If
    the full representation is needed, the methods get_full_vector() and
    get_full_matrix() can be used.
    """

    def __init__(
        self,
        objective: ObjectiveBase,
        lb: Union[np.ndarray, List[float]],
        ub: Union[np.ndarray, List[float]],
        dim_full: Optional[int] = None,
        x_fixed_indices: Optional[SupportsIntIterableOrValue] = None,
        x_fixed_vals: Optional[SupportsFloatIterableOrValue] = None,
        x_guesses: Optional[Iterable[float]] = None,
        startpoint_method: Optional[Callable] = None,
        x_names: Optional[Iterable[str]] = None,
        x_scales: Optional[Iterable[str]] = None,
        x_priors_defs: Optional[NegLogPriors] = None,
        lb_init: Union[np.ndarray, List[float], None] = None,
        ub_init: Union[np.ndarray, List[float], None] = None,
        copy_objective: bool = True,
    ):
        if copy_objective:
            objective = copy.deepcopy(objective)
        self.objective = objective

        self.lb_full: np.ndarray = np.array(lb).flatten()
        self.ub_full: np.ndarray = np.array(ub).flatten()
        if lb_init is None:
            lb_init = lb
        self.lb_init_full: np.ndarray = np.array(lb_init).flatten()
        if ub_init is None:
            ub_init = ub
        self.ub_init_full: np.ndarray = np.array(ub_init).flatten()

        self.dim_full: int = dim_full if dim_full is not None else \
            self.lb_full.size

        if x_fixed_indices is None:
            x_fixed_indices = []
        x_fixed_indices = _make_iterable_if_value(x_fixed_indices, 'int')
        self.x_fixed_indices: List[int] = [
            _type_conversion_with_check(idx, ix, 'fixed indices', 'int')
            for idx, ix in enumerate(x_fixed_indices)
        ]

        # We want the fixed values to be a list, since we might need to add
        # or remove values during profile computation
        if x_fixed_vals is None:
            x_fixed_vals = []
        x_fixed_vals = _make_iterable_if_value(x_fixed_vals, 'float')
        self.x_fixed_vals: List[float] = [
            _type_conversion_with_check(idx, x, 'fixed values', 'float')
            for idx, x in enumerate(x_fixed_vals)
        ]

        if x_guesses is None:
            x_guesses = np.zeros((0, self.dim_full))
        self.x_guesses_full: np.ndarray = np.array(x_guesses)

        self.startpoint_method = startpoint_method

        if x_names is None and objective.x_names is not None:
            x_names = objective.x_names
        elif x_names is None:
            x_names = [f'x{j}' for j in range(0, self.dim_full)]
        if len(set(x_names)) != len(x_names):
            raise ValueError("Parameter names x_names must be unique")
        self.x_names: List[str] = list(x_names)

        if x_scales is None:
            x_scales = ['lin'] * self.dim_full
        self.x_scales = x_scales

        self.x_priors = x_priors_defs

        self.normalize()

    @property
    def lb(self) -> np.ndarray:
        return self.lb_full[self.x_free_indices]

    @property
    def ub(self) -> np.ndarray:
        return self.ub_full[self.x_free_indices]

    @property
    def lb_init(self) -> np.ndarray:
        return self.lb_init_full[self.x_free_indices]

    @property
    def ub_init(self) -> np.ndarray:
        return self.ub_init_full[self.x_free_indices]

    @property
    def x_guesses(self) -> np.ndarray:
        return self.x_guesses_full[:, self.x_free_indices]

    @property
    def dim(self) -> int:
        return self.dim_full - len(self.x_fixed_indices)

    @property
    def x_free_indices(self) -> List[int]:
        return sorted(set(range(0, self.dim_full)) - set(self.x_fixed_indices))

    def normalize(self) -> None:
        """
        Reduce all vectors to dimension dim and have the objective accept
        vectors of dimension dim.
        """

        for attr in ['lb_full', 'lb_init_full', 'ub_full', 'ub_init_full']:
            value = self.__getattribute__(attr)
            if value.size == 1:
                self.__setattr__(attr, value * np.ones(self.dim_full))
            elif value.size == self.dim:
                # in this case the bounds only holds the values of the
                # reduced bounds.
                self.__setattr__(attr, self.get_full_vector(value,
                                                            self.x_fixed_vals))

            if self.__getattribute__(attr).size != self.dim_full:
                raise AssertionError(f"{attr} dimension invalid.")

        if self.x_guesses_full.shape[1] != self.dim_full:
            x_guesses_full = \
                np.empty((self.x_guesses_full.shape[0], self.dim_full))
            x_guesses_full[:] = np.nan
            x_guesses_full[:, self.x_free_indices] = self.x_guesses_full
            self.x_guesses_full = x_guesses_full

        # make objective aware of fixed parameters
        self.objective.update_from_problem(
            dim_full=self.dim_full,
            x_free_indices=self.x_free_indices,
            x_fixed_indices=self.x_fixed_indices,
            x_fixed_vals=self.x_fixed_vals)

        # sanity checks
        if len(self.x_scales) != self.dim_full:
            raise AssertionError("x_scales dimension invalid.")
        if len(self.x_names) != self.dim_full:
            raise AssertionError("x_names must be of length dim_full.")
        if len(self.x_fixed_indices) != len(self.x_fixed_vals):
            raise AssertionError(
                "x_fixed_indices and x_fixed_vals must have the same length."
            )
        if np.isnan(self.lb).any():
            raise ValueError('lb must not contain nan values')
        if np.isnan(self.ub).any():
            raise ValueError('ub must not contain nan values')
        if np.any(self.lb >= self.ub):
            raise ValueError('lb<ub not fulfilled.')

    def set_x_guesses(self,
                      x_guesses: Iterable[float]):
        """
        Sets the x_guesses of a problem.

        Parameters
        ----------
        x_guesses:
        """
        x_guesses_full = np.array(x_guesses)
        if x_guesses_full.shape[1] != self.dim_full:
            raise ValueError('The dimension of individual x_guesses must be '
                             'dim_full.')
        self.x_guesses_full = x_guesses_full

    def fix_parameters(self,
                       parameter_indices: SupportsIntIterableOrValue,
                       parameter_vals: SupportsFloatIterableOrValue) -> None:
        """
        Fix specified parameters to specified values
        """
        parameter_indices = _make_iterable_if_value(parameter_indices, 'int')
        parameter_vals = _make_iterable_if_value(parameter_vals, 'float')

        # first clean to-be-fixed indices to avoid redundancies
        for iter_index, (x_index, x_value) in enumerate(
                zip(parameter_indices, parameter_vals)
        ):
            # check if parameter was already fixed, otherwise add it to the
            # fixed parameters
            index = _type_conversion_with_check(iter_index, x_index,
                                                'indices', 'int')
            val = _type_conversion_with_check(iter_index, x_value,
                                              'values', 'float')
            if index in self.x_fixed_indices:
                self.x_fixed_vals[
                    self.x_fixed_indices.index(index)] = val
            else:
                self.x_fixed_indices.append(index)
                self.x_fixed_vals.append(val)

        self.normalize()

    def unfix_parameters(self, parameter_indices: SupportsIntIterableOrValue
                         ) -> None:
        """
        Free specified parameters
        """

        # check and adapt input
        parameter_indices = _make_iterable_if_value(parameter_indices, 'int')

        # first clean to-be-freed indices
        for iter_index, x_index in enumerate(parameter_indices):
            index = _type_conversion_with_check(iter_index, x_index,
                                                'indices', 'int')
            if index in self.x_fixed_indices:
                fixed_x_index = self.x_fixed_indices.index(index)
                self.x_fixed_indices.pop(fixed_x_index)
                self.x_fixed_vals.pop(fixed_x_index)

        self.normalize()

    def get_full_vector(
            self,
            x: Union[np.ndarray, None],
            x_fixed_vals: Iterable[float] = None
    ) -> Union[np.ndarray, None]:
        """
        Map vector from dim to dim_full. Usually used for x, grad.

        Parameters
        ----------
        x: array_like, shape=(dim,)
            The vector in dimension dim.
        x_fixed_vals: array_like, ndim=1, optional
            The values to be used for the fixed indices. If None, then nans are
            inserted. Usually, None will be used for grad and
            problem.x_fixed_vals for x.
        """
        if x is None:
            return None

        # make sure it is an array
        x = np.array(x)

        if len(x) == self.dim_full:
            return x

        # Note: The funny indexing construct is to handle residual gradients,
        # where the last dimension is assumed to be the parameter one.
        x_full = np.zeros(x.shape[:-1] + (self.dim_full,))
        x_full[:] = np.nan
        x_full[..., self.x_free_indices] = x
        if x_fixed_vals is not None:
            x_full[..., self.x_fixed_indices] = x_fixed_vals
        return x_full

    def get_full_matrix(
            self, x: Union[np.ndarray, None]
    ) -> Union[np.ndarray, None]:
        """
        Map matrix from dim to dim_full. Usually used for hessian.

        Parameters
        ----------
        x: array_like, shape=(dim, dim)
            The matrix in dimension dim.
        """
        if x is None:
            return None

        # make sure it is an array
        x = np.array(x)

        if len(x) == self.dim_full:
            return x

        x_full = np.zeros((self.dim_full, self.dim_full))
        x_full[:, :] = np.nan
        x_full[np.ix_(self.x_free_indices, self.x_free_indices)] = x

        return x_full

    def get_reduced_vector(
            self, x_full: Union[np.ndarray, None],
            x_indices: Optional[List[int]] = None
    ) -> Union[np.ndarray, None]:
        """
        Keep only those elements, which indices are specified in x_indices
        If x_indices is not provided, delete fixed indices.

        Parameters
        ----------
        x_full: array_like, ndim=1
            The vector in dimension dim_full.
        x_indices:
            indices of x_full that should remain
        """
        if x_full is None:
            return None

        if x_indices is None:
            x_indices = self.x_free_indices

        if len(x_full) == len(x_indices):
            return x_full

        x = [x_full[idx] for idx in x_indices]
        return np.array(x)

    def get_reduced_matrix(
            self, x_full: Union[np.ndarray, None]
    ) -> Union[np.ndarray, None]:
        """
        Map matrix from dim_full to dim, i.e. delete fixed indices.

        Parameters
        ----------
        x_full: array_like, ndim=2
            The matrix in dimension dim_full.
        """
        if x_full is None:
            return None

        if len(x_full) == self.dim:
            return x_full

        x = x_full[np.ix_(self.x_free_indices, self.x_free_indices)]

        return x

    def full_index_to_free_index(self, full_index: int):
        """Calculate index in reduced vector from index in full vector.

        Parameters
        ----------
        full_index: The index in the full vector.

        Returns
        -------
        free_index: The index in the free vector.
        """
        fixed_indices = np.asarray(self.x_fixed_indices)
        if full_index in fixed_indices:
            raise ValueError(
                "Cannot compute index in free vector: Index is fixed.")
        return full_index - sum(fixed_indices < full_index)

    def print_parameter_summary(self) -> None:
        """
        Prints a summary of what parameters are being optimized and
        parameter boundaries.
        """
        print(  # noqa: T001 (print)
            pd.DataFrame(
                index=self.x_names,
                data={
                    'free': [idx in self.x_free_indices
                             for idx in range(self.dim_full)],
                    'lb_full': self.lb_full,
                    'ub_full': self.ub_full
                }
            )
        )


_convtypes = {
    'float': {
        'attr': '__float__',
        'conv': float
    },
    'int': {
        'attr': '__int__',
        'conv': int
    }
}


def _type_conversion_with_check(index: int,
                                value: Union[SupportsFloat, SupportsInt],
                                valuename: str,
                                convtype: str) -> Union[float, int]:
    """
    Converts values to the requested type, raises and appropriate error if
    not possible.
    """

    if convtype not in _convtypes:
        raise ValueError(f'Unsupported type {convtype}')

    can_convert = hasattr(value, _convtypes[convtype]['attr'])
    # this may fail for weird custom ypes that can be converted to int but
    # not float, but we probably don't want those as indiced anyways
    lossless_conversion = not convtype == 'int' \
        or (hasattr(value, _convtypes['float']['attr'])
            and (float(value) - int(value) == 0.0))

    if not can_convert or not lossless_conversion:
        raise ValueError(
            f'All {valuename} must support lossless conversion to {convtype}. '
            f'Found type {type(value)} at index {index}, which cannot '
            f'be converted to {convtype}.'
        )

    return _convtypes[convtype]['conv'](value)


def _make_iterable_if_value(value: Union[SupportsFloatIterableOrValue,
                                         SupportsIntIterableOrValue],
                            convtype: str) -> Union[Iterable[SupportsFloat],
                                                    Iterable[SupportsInt]]:
    """
    Converts scalar values to iterables if input is scalar, may update type
    """

    if convtype not in _convtypes:
        raise ValueError(f'Unsupported type {convtype}')

    if not hasattr(value, '__iter__'):
        return [_type_conversion_with_check(0, value, 'values', convtype)]
    else:
        return value



import numpy as np
from typing import Union


def _check_none(fun):
    """Wrapper: Return None if any input argument is None."""
    def checked_fun(*args, **kwargs):
        if any(x is None for x in [*args, *(kwargs.values())]):
            return None
        return fun(*args, **kwargs)
    return checked_fun


@_check_none
def res_to_chi2(res: np.ndarray) -> Union[float, None]:
    """Translate residuals to chi2 values, `chi2 = sum(res**2)`."""
    return np.dot(res, res)


@_check_none
def chi2_to_fval(chi2: float) -> Union[float, None]:
    """Translate chi2 to function value, `fval = 0.5*chi2 = 0.5*sum(res**2)`.

    Note that for the function value we thus employ a probabilistic
    interpretation, as the log-likelihood of a standard normal noise model.
    This is in line with e.g. AMICI's and SciPy's objective definition.
    """
    return 0.5 * chi2


@_check_none
def res_to_fval(res: np.ndarray) -> Union[float, None]:
    """Translate residuals to function value, `fval = 0.5*sum(res**2)`."""
    return chi2_to_fval(res_to_chi2(res))


@_check_none
def sres_to_schi2(res: np.ndarray, sres: np.ndarray):
    """Translate residual sensitivities to chi2 gradient."""
    return 2 * res.dot(sres)


@_check_none
def schi2_to_grad(schi2: np.ndarray) -> np.ndarray:
    """Translate chi2 gradient to function value gradient.

    See also :func:`chi2_to_fval`.
    """
    return 0.5 * schi2


@_check_none
def sres_to_grad(res: np.ndarray, sres: np.ndarray):
    """Translate residual sensitivities to function value gradien, assuming
    `fval = 0.5*sum(res**2)`.

    See also :func:`chi2_to_fval`.
     """
    return schi2_to_grad(sres_to_schi2(res, sres))


@_check_none
def sres_to_fim(sres: np.ndarray):
    """Translate residual sensitivities to FIM.

    The FIM is based on the function values, not chi2, i.e. has a normalization
    of 0.5 as in :func:`res_to_fval`.
    """
    return sres.transpose().dot(sres)



import numpy as np

from ..objective import History
from ..problem import Problem


class OptimizerResult(dict):
    """
    The result of an optimizer run. Used as a standardized return value to
    map from the individual result objects returned by the employed
    optimizers to the format understood by pypesto.

    Can be used like a dict.

    Attributes
    ----------
    id:
        Id of the optimizer run. Usually the start index.
    x:
        The best found parameters.
    fval:
        The best found function value, `fun(x)`.
    grad:
        The gradient at `x`.
    hess:
        The Hessian at `x`.
    res:
        The residuals at `x`.
    sres:
        The residual sensitivities at `x`.
    n_fval
        Number of function evaluations.
    n_grad:
        Number of gradient evaluations.
    n_hess:
        Number of Hessian evaluations.
    n_res:
        Number of residuals evaluations.
    n_sres:
        Number of residual sensitivity evaluations.
    x0:
        The starting parameters.
    fval0:
        The starting function value, `fun(x0)`.
    history:
        Objective history.
    exitflag:
        The exitflag of the optimizer.
    time:
        Execution time.
    message: str
        Textual comment on the optimization result.

    Notes
    -----

    Any field not supported by the optimizer is filled with None.
    """

    def __init__(self,
                 id: str = None,
                 x: np.ndarray = None,
                 fval: float = None,
                 grad: np.ndarray = None,
                 hess: np.ndarray = None,
                 res: np.ndarray = None,
                 sres: np.ndarray = None,
                 n_fval: int = None,
                 n_grad: int = None,
                 n_hess: int = None,
                 n_res: int = None,
                 n_sres: int = None,
                 x0: np.ndarray = None,
                 fval0: float = None,
                 history: History = None,
                 exitflag: int = None,
                 time: float = None,
                 message: str = None):
        super().__init__()
        self.id = id
        self.x: np.ndarray = np.array(x) if x is not None else None
        self.fval: float = fval
        self.grad: np.ndarray = np.array(grad) if grad is not None else None
        self.hess: np.ndarray = np.array(hess) if hess is not None else None
        self.res: np.ndarray = np.ndarray(res) if res is not None else None
        self.sres: np.ndarray = np.ndarray(sres) if sres is not None else None
        self.n_fval: int = n_fval
        self.n_grad: int = n_grad
        self.n_hess: int = n_hess
        self.n_res: int = n_res
        self.n_sres: int = n_sres
        self.x0: np.ndarray = np.array(x0) if x0 is not None else None
        self.fval0: float = fval0
        self.history: History = history
        self.exitflag: int = exitflag
        self.time: float = time
        self.message: str = message

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def update_to_full(self, problem: Problem) -> None:
        """
        Updates values to full vectors/matrices

        Parameters
        ----------
        problem:
            problem which contains info about how to convert to full vectors
            or matrices
        """
        self.x = problem.get_full_vector(self.x, problem.x_fixed_vals)
        self.grad = problem.get_full_vector(self.grad)
        self.hess = problem.get_full_matrix(self.hess)
        self.x0 = problem.get_full_vector(self.x0, problem.x_fixed_vals)



import numpy as np
import warnings
from .clust_color import assign_colors
from .clust_color import assign_colors_for_list

from numbers import Number
from typing import Iterable, List, Optional, Union

from .constants import (
    LEN_RGB,
    LEN_RGBA,
    RGB,
    RGB_RGBA,
    RGBA_MIN,
    RGBA_MAX,
    RGBA_ALPHA,
    RGBA_WHITE,
)


def process_result_list(results, colors=None, legends=None):
    """
    assigns colors and legends to a list of results, check user provided lists

    Parameters
    ----------

    results: list or pypesto.Result
        list of pypesto.Result objects or a single pypesto.Result

    colors: list, optional
        list of RGBA colors

    legends: str or list
        labels for line plots

    Returns
    -------

    results: list of pypesto.Result
       list of pypesto.Result objects

    colors: list of RGBA
        One for each element in 'results'.

    legends: list of str
        labels for line plots
    """

    # check how many results were passed
    single_result = False
    legend_error = False
    if isinstance(results, list):
        if len(results) == 1:
            single_result = True
    else:
        single_result = True
        results = [results]

    # handle results according to their number
    if single_result:
        # assign colors and create list for later handling
        if colors is not None:
            colors = assign_colors(results, colors)
        colors = [colors]

        # create list of legends for later handling
        if not isinstance(legends, list):
            legends = [legends]
    else:
        # if more than one result is passed, we use one color per result
        colors = assign_colors_for_list(len(results), colors)

        # check whether list of legends has the correct length
        if legends is None:
            # No legends were passed: create some custom legends
            legends = []
            for i_leg in range(len(results)):
                legends.append('Result ' + str(i_leg))
        else:
            # legends were passed by user: check length
            if isinstance(legends, list):
                if len(legends) != len(results):
                    legend_error = True
            else:
                legend_error = True

    # size of legend list and size of results does not match
    if legend_error:
        raise ValueError('List of results passed and list of labels do '
                         'not have the same length but should. Stopping.')

    return results, colors, legends


def process_offset_y(offset_y: Optional[float],
                     scale_y: str,
                     min_val: float) -> float:
    """
    compute offset for y-axis, depend on user settings

    Parameters
    ----------

    offset_y:
       value for offsetting the later plotted values, in order to ensure
       positivity if a semilog-plot is used

    scale_y:
       Can be 'lin' or 'log10', specifying whether values should be plotted
       on linear or on log10-scale

    min_val:
        Smallest value to be plotted

    Returns
    -------

    offset_y: float
       value for offsetting the later plotted values, in order to ensure
       positivity if a semilog-plot is used
    """

    # check whether the offset specified by the user is sufficient
    if offset_y is not None:
        if (scale_y == 'log10') and (min_val + offset_y <= 0.):
            warnings.warn("Offset specified by user is insufficient. "
                          "Ignoring specified offset and using " +
                          str(np.abs(min_val) + 1.) + " instead.")
        else:
            return offset_y
    else:
        # check whether scaling is lin or log10
        if scale_y == 'lin':
            # linear scaling doesn't need any offset
            return 0.

    return 1. - min_val


def process_y_limits(ax, y_limits):
    """
    apply user specified limits of y-axis

    Parameters
    ----------

    ax: matplotlib.Axes, optional
        Axes object to use.

    y_limits: ndarray
       y_limits, minimum and maximum, for current axes object

    Returns
    -------

    ax: matplotlib.Axes, optional
        Axes object to use.
    """

    # apply y-limits, if they were specified by the user
    if y_limits is not None:
        y_limits = np.array(y_limits)

        # check validity of bounds
        if y_limits.size == 0:
            y_limits = np.array(ax.get_ylim())
        elif y_limits.size == 1:
            # if the user specified only an upper bound
            tmp_y_limits = ax.get_ylim()
            y_limits = [tmp_y_limits[0], y_limits]
        elif y_limits.size > 1:
            y_limits = [y_limits[0], y_limits[1]]

        # check validity of bounds if plotting in log-scale
        if ax.get_yscale() == 'log' and y_limits[0] <= 0.:
            tmp_y_limits = ax.get_ylim()
            if y_limits[1] <= 0.:
                y_limits = tmp_y_limits
                warnings.warn("Invalid bounds for plotting in "
                              "log-scale. Using defaults bounds.")
            else:
                y_limits = [tmp_y_limits[0], y_limits[1]]
                warnings.warn("Invalid lower bound for plotting in "
                              "log-scale. Using only upper bound.")

            # set limits
            ax.set_ylim(y_limits)

    else:
        # No limits passed, but if we have a result list: check the limits
        ax_limits = np.array(ax.get_ylim())
        data_limits = ax.dataLim.ymin, ax.dataLim.ymax

        # Check if data fits to axes and adapt limits, if necessary
        if ax_limits[0] > data_limits[0] or ax_limits[1] < data_limits[1]:
            # Get range of data
            data_range = data_limits[1] - data_limits[0]
            if ax.get_yscale() == 'log':
                data_range = np.log10(data_range)
                new_limits = (
                    np.power(10, np.log10(data_limits[0]) - 0.02 * data_range),
                    np.power(10, np.log10(data_limits[1]) + 0.02 * data_range))
            else:
                new_limits = (data_limits[0] - 0.02 * data_range,
                              data_limits[1] + 0.02 * data_range)

            # set limits
            ax.set_ylim(new_limits)

    return ax


def process_start_indices(start_indices: Union[int, Iterable[int]],
                          max_length: int) -> List[int]:
    """
    helper function that processes the start_indices and
    creates an array of indices if a number was provided and checks that the
    indices do not exceed the max_index

    Parameters
    ----------
    start_indices:
        list of indices or int specifying an endpoint of the sequence of
        indices
    max_length:
        maximum possible index for the start_indices
    """

    if isinstance(start_indices, Number):
        start_indices = range(int(start_indices))

    start_indices = np.array(start_indices, dtype=int)

    # check, whether index set is not too big
    start_indices = [start_index for start_index in start_indices if
                     start_index < max_length]

    return start_indices


def rgba2rgb(fg: RGB_RGBA, bg: RGB_RGBA = None) -> RGB:
    """Combine two colors, removing transparency.

    Parameters
    ----------
    fg:
        Foreground color.
    bg:
        Background color.

    Returns
    -------
    The combined color.
    """
    if bg is None:
        bg = RGBA_WHITE
    if len(bg) == LEN_RGBA:
        # return foreground if background is fully transparent
        if bg[RGBA_ALPHA] == RGBA_MIN:
            return fg
    else:
        if len(bg) != LEN_RGB:
            raise IndexError(
                'A background color of unexpected length was provided: {bg}'
            )
        bg = (*bg, RGBA_MAX)

    # return the foreground color if has no transparency
    if len(fg) == LEN_RGB or fg[RGBA_ALPHA] == RGBA_MAX:
        return fg
    if len(fg) != LEN_RGBA:
        raise IndexError(
            'A foreground color of unexpected length was provided: {fg}'
        )

    def apparent_composite_color_component(
            fg_component: float,
            bg_component: float,
            fg_alpha: float = fg[RGBA_ALPHA],
            bg_alpha: float = bg[RGBA_ALPHA],
    ) -> float:
        """
        Composite a foreground color component over a background color
        component.

        Porter and Duff equations are used for alpha compositing.

        Parameters
        ----------
        fg_component:
            The foreground color component.
        bg_component:
            The background color component.
        fg_alpha:
            The foreground color transparency/alpha component.
        bg_alpha:
            The background color transparency/alpha component.

        Returns
        -------
        The component of the new color.
        """
        return (
            fg_component * fg_alpha +
            bg_component * bg_alpha * (RGBA_MAX - fg_alpha)
        ) / (fg_alpha + bg_alpha * (RGBA_MAX - fg_alpha))

    return [
        apparent_composite_color_component(fg[i], bg[i])
        for i in range(LEN_RGB)
    ]



import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
from .reference_points import create_references, ReferencePoint
from .clust_color import assign_colors, delete_nan_inf, RGBA
from .misc import (
    process_result_list, process_start_indices, process_y_limits,
    process_offset_y
)

from pypesto import Result
from typing import Optional, Union, Sequence, Tuple, List


def waterfall(results: Union[Result, Sequence[Result]],
              ax: Optional[plt.Axes] = None,
              size: Optional[Tuple[float]] = (18.5, 10.5),
              y_limits: Optional[Tuple[float]] = None,
              scale_y: Optional[str] = 'log10',
              offset_y: Optional[float] = None,
              start_indices: Optional[Union[Sequence[int], int]] = None,
              reference: Optional[Sequence[ReferencePoint]] = None,
              colors: Optional[Union[RGBA, Sequence[RGBA]]] = None,
              legends: Optional[Union[Sequence[str], str]] = None):
    """
    Plot waterfall plot.

    Parameters
    ----------

    results:
        Optimization result obtained by 'optimize.py' or list of those

    ax: matplotlib.Axes, optional
        Axes object to use.

    size:
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified

    y_limits: float or ndarray, optional
        maximum value to be plotted on the y-axis, or y-limits

    scale_y:
        May be logarithmic or linear ('log10' or 'lin')

    offset_y:
        offset for the y-axis, if it is supposed to be in log10-scale

    start_indices:
        Integers specifying the multistart to be plotted or int specifying
        up to which start index should be plotted

    reference:
        Reference points for optimization results, containing at least a
        function value fval

    colors:
        Colors or single color  for plotting. If not set, clustering is done
        and colors are assigned automatically

    legends:
        Labels for line plots, one label per result object

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # parse input
    (results, colors, legends) = process_result_list(results, colors, legends)

    refs = create_references(references=reference)

    # precompute y-offset, if needed and if a list of results was passed
    fvals_all, offset_y = process_offset_for_list(offset_y, results, scale_y,
                                                  start_indices, refs)

    # plotting routine needs the maximum number of multistarts
    max_len_fvals = np.array([0])

    # loop over results
    for j, fvals in enumerate(fvals_all):
        # extract specific cost function values from result
        max_len_fvals = np.max([max_len_fvals, *fvals.shape])

        # call lowlevel plot routine
        ax = waterfall_lowlevel(fvals=fvals, scale_y=scale_y,
                                offset_y=offset_y, ax=ax, size=size,
                                colors=colors[j], legend_text=legends[j])

    # apply changes specified be the user to the axis object
    ax = handle_options(ax, max_len_fvals, refs, y_limits, offset_y)
    return ax


def waterfall_lowlevel(fvals, scale_y='log10', offset_y=0., ax=None,
                       size=(18.5, 10.5), colors=None, legend_text=None):
    """
    Plot waterfall plot using list of function values.

    Parameters
    ----------

    fvals: numeric list or array
        Including values need to be plotted.

    scale_y: str, optional
        May be logarithmic or linear ('log10' or 'lin')

    offset_y:
        offset for the y-axis, if it is supposed to be in log10-scale

    ax: matplotlib.Axes, optional
        Axes object to use.

    size: tuple, optional
        see waterfall

    colors: list, or RGBA, optional
        list of colors, or single color
        color or list of colors for plotting. If not set, clustering is done
        and colors are assigned automatically

    legend_text: str
        Label for line plots

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # axes
    if ax is None:
        ax = plt.subplots()[1]
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # parse input
    fvals = np.array(fvals)
    # remove nan or inf values in fvals
    _, fvals = delete_nan_inf(fvals)

    fvals.sort()
    n_fvals = len(fvals)
    start_ind = range(n_fvals)

    # assign colors
    # note: this has to happen before sorting
    # to get the same colors in different plots
    colors = assign_colors(fvals, colors=colors)

    # plot
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    # plot line
    if scale_y == 'log10':
        ax.semilogy(start_ind, fvals, color=[0.7, 0.7, 0.7, 0.6])
    else:
        ax.plot(start_ind, fvals, color=[0.7, 0.7, 0.7, 0.6])

    # plot points
    for j in range(n_fvals):
        # parse data for plotting
        color = colors[j]
        fval = fvals[j]
        if j == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # line plot (linear or logarithmic)
        if scale_y == 'log10':
            ax.semilogy(j, fval, color=color,
                        marker='o', label=tmp_legend, alpha=1.)
        else:
            ax.plot(j, fval, color=color,
                    marker='o', label=tmp_legend, alpha=1.)

    # check if y-axis has a reasonable scale
    y_min, y_max = ax.get_ylim()
    if scale_y == 'log10':
        if np.log10(y_max) - np.log10(y_min) < 1.:
            y_mean = 0.5 * (np.log10(y_min) + np.log10(y_max))
            ax.set_ylim(10. ** (y_mean - 0.5), 10. ** (y_mean + 0.5))
    else:
        if y_max - y_min < 1.:
            y_mean = 0.5 * (y_min + y_max)
            ax.set_ylim(y_mean - 0.5, y_mean + 0.5)

    # labels
    ax.set_xlabel('Ordered optimizer run')
    if offset_y == .0:
        ax.set_ylabel('Function value')
    else:
        ax.set_ylabel('Offsetted function value (relative to best start)')
    ax.set_title('Waterfall plot')
    if legend_text is not None:
        ax.legend()

    return ax


def process_offset_for_list(
        offset_y: float,
        results: Sequence[Result],
        scale_y: Optional[str],
        start_indices: Optional[Sequence[int]] = None,
        references: Optional[Sequence[ReferencePoint]] = None,
) -> Tuple[List[np.ndarray], float]:
    """
    If we have a list of results, all should use the same offset_y,
    which is computed by this function and added to the fvals

    Parameters
    ----------

    offset_y:
        User provided offset_y

    results:
        Optimization results obtained by 'optimize.py'

    scale_y:
        May be logarithmic or linear ('log10' or 'lin')

    start_indices:
        Integers specifying the multistart to be plotted or int specifying
        up to which start index should be plotted


    references:
        Reference points that will be plotted along with the results

    Returns
    -------

    fvals:
        List of arrays of function values for each result

    offset_y:
        offset for the y-axis
    """
    min_val = np.inf
    fvals_all = []
    for result in results:
        fvals = np.asarray([
            np.array(result.optimize_result.get_for_key('fval'))
        ])
        if start_indices is None:
            start_indices = np.array(range(fvals.size))
        else:
            start_indices = process_start_indices(start_indices, fvals.size)
        fvals = fvals[:, start_indices]
        # if none of the fvals are finite, set default value to zero as
        # np.nanmin will error for an empty array
        if np.isfinite(fvals).any():
            min_val = min(min_val, np.nanmin(fvals[np.isfinite(fvals)]))

        fvals_all.append(fvals)

    # if there are references, also account for those
    if references:
        min_val = min(min_val, np.nanmin([r['fval'] for r in references]))

    offset_y = process_offset_y(offset_y, scale_y, float(min_val))

    # return offsetted values
    return [fvals + offset_y for fvals in fvals_all], offset_y


def handle_options(ax, max_len_fvals, ref, y_limits, offset_y):
    """
    Get the limits for the y-axis, plots the reference points, will do
    more at a later time point. This function is there to apply whatever
    kind of post-plotting transformations to the axis object.

    Parameters
    ----------

    ax: matplotlib.Axes, optional
        Axes object to use.

    max_len_fvals: int
        maximum number of points

    ref: list, optional
        List of reference points for optimization results, containing at
        least a function value fval

    y_limits: float or ndarray, optional
        maximum value to be plotted on the y-axis, or y-limits

    offset_y:
        offset for the y-axis, if it is supposed to be in log10-scale

    Returns
    -------

    ax: matplotlib.Axes
        The plot axes.
    """

    # handle reference points
    for i_ref in ref:
        # plot reference point as line
        ax.plot([0, max_len_fvals - 1],
                [i_ref.fval + offset_y, i_ref.fval + offset_y], '--',
                color=i_ref.color, label=i_ref.legend)

        # create legend for reference points
        if i_ref.legend is not None:
            ax.legend()

    # handle y-limits
    ax = process_y_limits(ax, y_limits)

    return ax



"""
Storage
=======

Saving and loading traces and results objects.
"""

from .save_to_hdf5 import (ProblemHDF5Writer, OptimizationResultHDF5Writer,
                           ProfileResultHDF5Writer, SamplingResultHDF5Writer,
                           write_result, get_or_create_group)
from .read_from_hdf5 import (ProblemHDF5Reader, OptimizationResultHDF5Reader,
                             ProfileResultHDF5Reader, SamplingResultHDF5Reader,
                             read_result, load_objective_config,
                             optimization_result_from_history)
from .hdf5 import write_array
